@@ -1,23 +1,13 @@
 /**
  * @file Beagle.java
  *
- * Copyright 2009-2016 Phylogenetic Likelihood Working Group
+ * Copyright 2009-2019 Phylogenetic Likelihood Working Group
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @brief This file documents the API as well as header for the
  * Broad-platform Evolutionary Analysis General Likelihood Evaluator
@@ -144,7 +134,7 @@ public interface Beagle extends Serializable {
      * CPU implementation. It should only be called after beagleCreateInstance and
      * requires the THREADING_CPP flag to be set. It has no effect on GPU-based
      * implementations. It has no effect with the default THREADING_NONE setting.
-     * If THREADING_CPP is set and this function is not called BEAGLE will use 
+     * If THREADING_CPP is set and this function is not called BEAGLE will use
      * a heuristic to set an appropriate number of threads.
      *
      * @param threadCount          Number of threads (input)
@@ -213,6 +203,23 @@ public interface Beagle extends Serializable {
             int tipIndex,
             final double[] inPartials);
 
+    /**
+     * Set the pre-order partials for the root node
+     *
+     * This function copies an array of stateFrequencies into an instance buffer as the
+     * pre-order partials for the root node.
+     *
+     * @param inbufferIndices                List of partialsBuffer indices to set (input)
+     * @param instateFrequenciesIndices      List of state frequencies for each partialsBuffer (input).
+     *                                       There should be one set for each of parentBufferIndices
+     * @param count                          Number of partialsBuffer to integrate(input)
+     */
+    void setRootPrePartials(
+            final int[] inbufferIndices,
+            final int[] instateFrequenciesIndices,
+            int count
+    );
+
     /**
      * Set an instance partials buffer
      *
@@ -240,19 +247,19 @@ public interface Beagle extends Serializable {
             int bufferIndex,
             int scaleIndex,
             final double[] outPartials);
-                        
+
     /**
      * Get scale factors from instance buffer on log-scale
      *
      * This function copies an array of scale factors from an instance buffer. The outFactors array should
      * be patternCount in length.
-     *   
+     *
      * @param scaleIndex    Index of scaleBuffer to get (input)
      * @param  outFactors  Pointer to which to receive partialsBuffer (output)
      */
-    void getLogScaleFactors(           
+    void getLogScaleFactors(
             int scaleIndex,
-            final double[] outFactors);            
+            final double[] outFactors);
 
     /**
      * Set an eigen-decomposition buffer
@@ -325,7 +332,38 @@ public interface Beagle extends Serializable {
             final int[] secondIndices,
             final int[] resultIndices,
             int matrixCount);
-    
+
+
+    /**
+     * Add lists of transition probability matrices
+     *
+     * This function add two lists of transition probability matrices.
+     *
+     * @param firstIndices              List of indices of the first transition probability matrices to convolve (input)
+     * @param secondIndices             List of indices of the second transition probability matrices to convolve (input)
+     * @param resultIndices             List of indices of resulting transition probability matrices (input)
+     * @param matrixCount               Lenght of lists
+     */
+    void addTransitionMatrices(
+            final int[] firstIndices,
+            final int[] secondIndices,
+            final int[] resultIndices,
+            int matrixCount);
+
+    /**
+     * Transpose lists of transition probability matrices
+     *
+     * This function transposes a lists of transition probability matrices.
+     *
+     * @param inIndices              List of indices of the transition probability matrices to transpose (input)
+     * @param outIndices             List of indices of the resulting transition probability matrices (input)
+     * @param matrixCount               Lenght of lists
+     */
+    void transposeTransitionMatrices(
+    		final int[] inIndices,
+    		final int[] outIndices,
+    		int matrixCount);
+
     /**
      * Calculate a list of transition probability matrices
      *
@@ -378,7 +416,7 @@ public interface Beagle extends Serializable {
      *
      * This function copies a finite-time transition probability matrix into a matrix buffer.
      * @param matrixIndex   Index of matrix buffer (input)
-     * @param inMatrix          Pointer to source transition probability matrix (input)
+     * @param inMatrix      Pointer to source transition probability matrix (input)
      * @param paddedValue   Value to be used for padding for ambiguous states (e.g. 1 for probability matrices, 0 for derivative matrices) (input)
      */
     void setTransitionMatrix(
@@ -386,6 +424,19 @@ public interface Beagle extends Serializable {
             final double[] inMatrix, 	/**< Pointer to source transition probability matrix (input) */
             double paddedValue);
 
+    /**
+     * This function copies a differential transition probability matrix into a matrix buffer.
+     * The inMatrix array should be of size stateCount * stateCount * categoryCount and will
+     * contain one matrix for each rate category.
+     *
+     * This function copies a differential transition probability matrix into a matrix buffer.
+     * @param matrixIndex   Index of matrix buffer (input)
+     * @param inMatrix      Pointer to source transition probability matrix (input)
+     */
+    void setDifferentialMatrix(
+            int matrixIndex,			/**< Index of matrix buffer (input) */
+            final double[] inMatrix); 	/**< Pointer to source transition probability matrix (input) */
+
     /**
      * Get a finite-time transition probability matrix
      *
@@ -400,6 +451,119 @@ public interface Beagle extends Serializable {
     void getTransitionMatrix(int matrixIndex,
                              double[] outMatrix);
 
+    /**
+     * Calculate or queue for calculation pre-order partials using a list of operations
+     *
+     * This function either calculates or queues for calculation a list pre-order partials. Implementations
+     * supporting ASYNCH may queue these calculations while other implementations perform these
+     * operations immediately and in order.
+     *
+     * Operations list is a list of 7-tuple integer indices, with one 7-tuple per operation.
+     * Format of 7-tuple operation: {destinationPartials,
+     *                               destinationScaleWrite,
+     *                               destinationScaleRead,
+     *                               pre-order partials of the parent node,
+     *                               Ptr matrices of the current node,
+     *                               post-order partials of the sibling node,
+     *                               Ptr matrices of the sibling node}
+     *
+     * ///TODO: consider > 3 degree nodes?
+     *
+     * @param operations                list of 7-tuples specifying operations (input)
+     * @param operationCount            Number of operations (input)
+     * @param cumulativeScaleIndex      Index number of scaleBuffer to store accumulated factors (input)
+     *
+     */
+    void updatePrePartials(
+            final int[] operations,
+            int operationCount,
+            int cumulativeScaleIndex);
+
+    void updatePrePartialsByPartition(
+            final int[] operations,
+            int operationCount);
+
+    /**
+     * Calculate gradient or / and diagonal hessian of given edges
+     *
+     * This function calculates gradient or diagonal hessian of the log likelihood with respect to edge length.
+     *
+     * @param postBufferIndices                 list of post order buffer indices
+     * @param preBufferIndices                  list of pre  order buffer indices
+     * @param rootBufferIndex                   root post order buffer index
+     * @param firstDerivativeIndices            Q matrix indices
+     * @param secondDerivativeIndices           Q^2 matrix indices
+     * @param categoryWeightsIndex              category weights index
+     * @param categoryRatesIndex                category rates index
+     * @param stateFrequenciesIndex             state frequency index
+     * @param cumulativeScaleIndices            cumulative scaling factor indices, currently not used
+     * @param count                             number of edges
+     * @param outFirstDerivative                gradient output array
+     * @param outDiagonalSecondDerivative       diagonal hessian output array
+     *
+     */
+    void calculateEdgeDerivative(
+    		final int[] postBufferIndices,
+	        final int[] preBufferIndices,
+	        final int rootBufferIndex,
+    	    final int[] firstDerivativeIndices,
+	        final int[] secondDerivativeIndices,
+	        final int categoryWeightsIndex,
+	        final int categoryRatesIndex,
+	        final int stateFrequenciesIndex,
+	        final int[] cumulativeScaleIndices,
+	        int count,
+	        double[] outFirstDerivative,
+    	    double[] outDiagonalSecondDerivative);
+
+    /**
+     * Calculate differential w.r.t. edges
+     *
+     * This function calculates a derivative of the log likelihood with respect to edge-differentials.
+     *
+     * @param postBufferIndices                 list of post order buffer indices
+     * @param preBufferIndices                  list of pre  order buffer indices
+     * @param derivativeMatrixIndices           differential Q matrix indices
+     * @param categoryWeightsIndices            category weights indices
+     * @param count                             number of edges
+     * @param outDerivatives                    derivative-per-site output array
+     * @param outSumDerivatives                 sum of derivatives across sites output array
+     * @param outSumSquaredDerivatives          sum of squared derivatives output array
+     *
+     */
+    void calculateEdgeDifferentials(
+    		final int[] postBufferIndices,
+        	final int[] preBufferIndices,
+	        final int[] derivativeMatrixIndices,
+    	    final int[] categoryWeightsIndices,
+        	int count,
+	        double[] outDerivatives,
+	        double[] outSumDerivatives,
+	        double[] outSumSquaredDerivatives);
+
+    /**
+     * Calculate differential w.r.t. substitution-model generator elements
+     *
+     * This function calculates a derivative of the log likelihood with respect to the
+     * substitution model generator elements
+     *
+     * @param postBufferIndices                 list of post order buffer indices
+     * @param preBufferIndices                  list of pre  order buffer indices
+     * @param categoryWeightsIndices            category weights indices
+     * @param count                             number of edges
+     * @param outDerivatives                    derivative-per-element output array
+     *
+     */
+    void calculateCrossProductDifferentials(
+    		final int[] postBufferIndices,
+        	final int[] preBufferIndices,
+    	    final int[] categoryRateIndices,
+    	    final int[] categoryWeightsIndices,
+        	final double[] edgeLengths,
+        	int count,
+	        double[] outSumDerivatives,
+	        double[] outSumSquaredDerivatives);
+
     /**
      * Calculate or queue for calculation partials using a list of operations
      *
@@ -537,7 +701,7 @@ public interface Beagle extends Serializable {
     void copyScaleFactors(
         int destScalingIndex,
         int srcScalingIndex
-    );    
+    );
 
     /**
      * Reset scalefactors
@@ -597,8 +761,8 @@ public interface Beagle extends Serializable {
      *                                      should be one set for each of parentBufferIndices
      * @param cumulativeScaleIndices    List of scalingFactors indices to accumulate over (input). There
      *                                      should be one set for each of parentBufferIndices
-     * @param partitionIndices          List of partition indices indicating which sites in each 
-     *                                  partialsBuffer should be used (input). There should be one 
+     * @param partitionIndices          List of partition indices indicating which sites in each
+     *                                  partialsBuffer should be used (input). There should be one
      *                                  index for each of bufferIndices
      * @param partitionCount            Number of partialsBuffer to integrate (input)
      * @param count                     Number of sets of partitions to integrate across (input)
@@ -664,4 +828,4 @@ public interface Beagle extends Serializable {
      * @return
      */
     public InstanceDetails getDetails();
-}
\ No newline at end of file
+}
