@@ -3,7 +3,7 @@ package beagle;
 public class GeneralBeagleImpl implements Beagle {
 
     public static final boolean DEBUG = false;
-    public static final boolean SCALING = true;
+    public static final boolean SCALING = true;  //this was true XJ
 
     // These settings are chosen for single precision computation.
     // The single precision exponents go from -126 to 127 (2 ^ x)
@@ -177,6 +177,23 @@ public class GeneralBeagleImpl implements Beagle {
         }
     }
 
+    public void setRootPrePartials(int[] inbufferIndices, int[] instateFrequenciesIndices, int count){
+        if(count == 1){
+            int stateFrequenciesIndex = instateFrequenciesIndices[0];
+            int bufferIndex = inbufferIndices[0];
+            if (this.partials[bufferIndex] == null){
+                this.partials[bufferIndex] = new double[partialsSize];
+            }
+            int k = 0;
+            for(int i = 0; i < categoryCount; i++){
+                for(int j = 0; j < patternCount; j++){
+                    System.arraycopy(this.stateFrequencies[stateFrequenciesIndex], 0, this.partials[bufferIndex], k, stateCount);
+                    k += stateCount;
+                }
+            }
+        }
+    }
+
     public void setPartials(final int bufferIndex, final double[] partials) {
         assert(this.partials[bufferIndex] != null);
         System.arraycopy(partials, 0, this.partials[bufferIndex], 0, partialsSize);
@@ -228,11 +245,14 @@ public class GeneralBeagleImpl implements Beagle {
         System.arraycopy(inMatrix, 0, this.matrices[matrixIndex], 0, this.matrixSize);
     }
 
+    public void setDifferentialMatrix(final int matrixIndex, final double[] inMatrix) {
+    	setTransitionMatrix(matrixIndex, inMatrix, 0.0);
+    }
+
     public void getTransitionMatrix(final int matrixIndex, final double[] outMatrix) {
         System.arraycopy(this.matrices[matrixIndex],0,outMatrix,0,outMatrix.length);
     }
 
-    
 	// /////////////////////////
 	// ---TODO: Epoch model---//
 	// /////////////////////////
@@ -243,32 +263,32 @@ public class GeneralBeagleImpl implements Beagle {
             int matrixCount) {
 
 		for (int u = 0; u < matrixCount; u++) {
-			
+
 			  int firstIndex = firstIndices[u];
 			  int secondIndex = secondIndices[u];
 			  int resultIndex = resultIndices[u];
-		
+
 			     if (DEBUG) {
 			    	 System.err.println("Convolving matrix " + firstIndex + " with matrix  " + secondIndex + " into " + resultIndex);
 			    	 }
-			  
+
 			for (int l = 0; l < categoryCount; l++) {
 
 				for (int i = 0; i < stateCount; i++) {
 					for (int j = 0; j < stateCount; j++) {
 
-						
+
 						for (int k = 0; k < stateCount; k++) {
-							
-//							sum += 
-							
+
+//							sum +=
+
 						}//END: k loop
-						
-						
-						
-						
-						
-						
+
+
+
+
+
+
 					}// END: j loop
 				}// END: i loop
 			}// /END: l loop
@@ -276,8 +296,21 @@ public class GeneralBeagleImpl implements Beagle {
 		}// END: u loop
 
 	}// END: convolveTransitionMatrices
-    
-    
+
+    public void addTransitionMatrices(
+            final int[] firstIndices,
+            final int[] secondIndices,
+            final int[] resultIndices,
+            int matrixCount) {
+        throw new RuntimeException("Not yet implemented");
+    }
+
+	public void transposeTransitionMatrices(final int[] inIndices,
+											final int[] outIndices,
+											int matrixCount) {
+		throw new RuntimeException("Not yet implemented");
+	}
+
     public void updateTransitionMatrices(final int eigenIndex,
                                          final int[] probabilityIndices,
                                          final int[] firstDerivativeIndices,
@@ -385,6 +418,85 @@ public class GeneralBeagleImpl implements Beagle {
         }
     }
 
+    /**
+     * Operations list is a list of 7-tuple integer indices, with one 7-tuple per operation.
+     * Format of 7-tuple operation: {destinationPartials,
+     *                               destinationScaleWrite,
+     *                               destinationScaleRead,
+     *                               pre-order partials of the parent node,
+     *                               Ptr matrices of the current node,
+     *                               post-order partials of the sibling node,
+     *                               Ptr matrices of the sibling node}
+     *
+     */
+    public void updatePrePartials(final int[] operations, final int operationCount, final int cumulativeScaleIndex) {
+
+        int x = 0;
+        for (int op = 0; op < operationCount; op++) {
+            int bufferIndex3 = operations[x];
+            int bufferIndex1 = operations[x + 3];
+            int matrixIndex1 = operations[x + 4];
+            int bufferIndex2 = operations[x + 5];
+            int matrixIndex2 = operations[x + 6];
+
+            x += Beagle.OPERATION_TUPLE_SIZE;
+
+            int exponent = 0;
+
+            exponent = updatePrePartialsPartials(bufferIndex1, matrixIndex1, bufferIndex2, matrixIndex2, bufferIndex3);
+
+            if (SCALING) {
+                if (exponent > SCALING_EXPONENT_THRESHOLD) {
+                    rescalePartials(bufferIndex3);
+                }
+            }
+        }
+    }
+
+    public void updatePrePartialsByPartition(final int[] operations, final int operationCount) {
+        throw new BeagleException("updatePrePartialsByPartition", -1);
+    }
+
+	@Override
+    public void calculateEdgeDifferentials(final int[] postBufferIndices,
+                                           final int[] preBufferIndices,
+                                           final int[] derivativeMatrixIndices,
+                                           final int[] categoryWeightsIndices,
+                                           int count,
+                                           double[] outDerivatives,
+                                           double[] outSumDerivatives,
+                                           double[] outSumSquaredDerivatives) {
+		throw new RuntimeException("Not yet implemented");
+    }
+
+	@Override
+    public void calculateCrossProductDifferentials(final int[] postBufferIndices,
+                                           		   final int[] preBufferIndices,
+                                           		   final int[] categoryRateIndices,
+                                           		   final int[] categoryWeightsIndices,
+                                           		   final double[] edgeLengths,
+                                           		   int count,
+                                           		   double[] outSumDerivatives,
+                                           		   double[] outSumSquaredDerivatives) {
+		throw new RuntimeException("Not yet implemented");
+    }
+
+    @Override
+    public void calculateEdgeDerivative(int[] postBufferIndices,
+    									int[] preBufferIndices,
+    									int rootBufferIndex,
+    									int[] firstDerivativeIndices,
+    									int[] secondDerivativeIndices,
+    									int categoryWeightsIndex,
+    									int categoryRatesIndex,
+    									int stateFrequenciesIndex,
+    									int[] cumulativeScaleIndices,
+    									int count,
+    									double[] outFirstDerivative,
+    									double[] outDiagonalSecondDerivative) {
+        throw new RuntimeException("Not yet implemented");
+    }
+
     /**
      * Operations list is a list of 9-tuple integer indices, with one 7-tuple per operation.
      * Format of 9-tuple operation: {destinationPartials,
@@ -721,6 +833,86 @@ public class GeneralBeagleImpl implements Beagle {
         return exponent;
     }
 
+    protected int updatePrePartialsPartials(int bufferIndex1, int matrixIndex1, int bufferIndex2, int matrixIndex2, int bufferIndex3)
+    {
+        double[] matrices1 = matrices[matrixIndex1];  // Ptr matrices of the current node
+        double[] matrices2 = matrices[matrixIndex2];  // Ptr matrices of the sibling node
+
+        double[] partials1 = partials[bufferIndex1];  // pre-order partials of the parent node
+
+        double[] partials2 = partials[bufferIndex2];  // post-order partials of the sibling node
+
+        double[] partials3 = partials[bufferIndex3];  //destPartials
+
+        double sum2, MjPj;
+
+        int exponent = 0;
+        if (SCALING) {
+            int[] counts1 = scalingFactorCounts[bufferIndex1];
+            int[] counts2 = scalingFactorCounts[bufferIndex2];
+            int[] counts3 = scalingFactorCounts[bufferIndex3];
+            for (int i = 0; i < counts1.length; i++) {
+                // The scale factor counts is the sum of the two nodes below
+                counts3[i] = counts1[i] + counts2[i];
+            }
+        }
+
+        int u = 0;
+        int v = 0;
+        for (int l = 0; l < categoryCount; l++) {
+
+            for (int k = 0; k < patternCount; k++) {
+
+                int w = l * matrixSize;
+
+
+                for (int i = 0; i < stateCount; i++) {
+
+                    sum2 = 0.0;
+
+                    for (int j = 0; j < stateCount; j++) {
+                        sum2 += matrices2[w] * partials2[v + j];
+                        w++;
+                    }
+
+                    // sum2 = M_j P_j
+                    // now 2nd loop
+
+                    MjPj = sum2 * partials1[v + i];
+
+                    w = l * matrixSize;
+                    for (int j = 0; j < stateCount; j++){
+                        partials3[v + j] += matrices1[w] * MjPj;
+                        w++;
+                    }
+
+//                    if (SCALING) {
+//                        // this is to find the largest absolute exponent
+//                        exponent |= Math.abs(Math.getExponent(partials3[u]));
+//                    }
+
+                    u++;
+                }
+                v += stateCount;
+
+            }
+
+            if (DEBUG) {
+//    	    	System.err.println("1:PP node = "+nodeIndex3);
+//    	    	for(int p=0; p<partials3.length; p++) {
+//    	    		System.err.println("1:PP\t"+partials3[p]);
+//    	    	}
+//                System.err.println("node = "+nodeIndex3);
+//                System.err.println(new dr.math.matrixAlgebra.Vector(partials3));
+//                System.err.println(new dr.math.matrixAlgebra.Vector(scalingFactors[currentPartialsIndices[nodeIndex3]][nodeIndex3]));
+                //System.exit(-1);
+            }
+        }
+
+        return exponent;
+    }
+
+
     public void calculateRootLogLikelihoods(final int[] bufferIndices, final int[] categoryWeightsIndices, final int[] stateFrequenciesIndices, final int[] cumulativeScaleIndices, final int count, final double[] outSumLogLikelihood) {
 
         assert(count == 1); // @todo implement integration across multiple subtrees
@@ -798,4 +990,4 @@ public class GeneralBeagleImpl implements Beagle {
         return details;
     }
 
-}
\ No newline at end of file
+}
