@@ -6,19 +6,9 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @author Marc Suchard
  * @author David Swofford
@@ -55,10 +45,34 @@
 		V_Real tmp_##dest##01, tmp_##dest##23; \
 		tmp_##dest##01 = _mm_load_pd(&src[v + 0]); \
 		tmp_##dest##23 = _mm_load_pd(&src[v + 2]); \
-		dest##0 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(0,0)); \
-		dest##1 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(1,1)); \
-		dest##2 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(0,0)); \
-		dest##3 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(1,1));
+		dest##0 = VEC_SHUFFLE0(tmp_##dest##01, tmp_##dest##01); \
+		dest##1 = VEC_SHUFFLE1(tmp_##dest##01, tmp_##dest##01); \
+		dest##2 = VEC_SHUFFLE0(tmp_##dest##23, tmp_##dest##23); \
+		dest##3 = VEC_SHUFFLE1(tmp_##dest##23, tmp_##dest##23);
+//  	dest##0 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(0,0)); \
+// 	 	dest##1 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(1,1)); \
+// 	 	dest##2 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(0,0)); \
+// 		dest##3 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(1,1));
+
+#define SSE_PREFETCH_VECTORIZED_PARTIALS(dest, src, v) \
+        dest##01 = _mm_load_pd(&src[v + 0]); \
+        dest##23 = _mm_load_pd(&src[v + 2]);
+
+#define SSE_VECTORIZED_INNER_PRODUCT(lhs, rhs) \
+        VEC_ADD(_mm_dp_pd(lhs##01, rhs##01, 0xff), _mm_dp_pd(lhs##23, rhs##23, 0xff))
+
+#define SSE_SCHUR_PRODUCT_PARTIALS(dest, src, v, srcq) \
+		V_Real tmp_##dest##01, tmp_##dest##23; \
+		tmp_##dest##01 = VEC_MULT(_mm_load_pd(&src[v + 0]), srcq##01); \
+		tmp_##dest##23 = VEC_MULT(_mm_load_pd(&src[v + 2]), srcq##23); \
+		dest##0 = VEC_SHUFFLE0(tmp_##dest##01, tmp_##dest##01); \
+		dest##1 = VEC_SHUFFLE1(tmp_##dest##01, tmp_##dest##01); \
+		dest##2 = VEC_SHUFFLE0(tmp_##dest##23, tmp_##dest##23); \
+		dest##3 = VEC_SHUFFLE1(tmp_##dest##23, tmp_##dest##23);
+// 		dest##0 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(0,0)); \
+// 		dest##1 = _mm_shuffle_pd(tmp_##dest##01, tmp_##dest##01, _MM_SHUFFLE2(1,1)); \
+// 		dest##2 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(0,0)); \
+// 		dest##3 = _mm_shuffle_pd(tmp_##dest##23, tmp_##dest##23, _MM_SHUFFLE2(1,1));
 #endif
 
 /* Loads (transposed) finite-time transition matrices into SSE vectors */
@@ -76,6 +90,21 @@
 		dest_vu_m2[i][1].x[1] = m2[3*OFFSET]; \
 	}
 
+/* Loads (only transpose matrix 2) finite-time transition matrices into SSE vectors */
+#define SSE_PREFETCH_PRE_MATRICES(src_m1, src_m2, dest_vu_m1, dest_vu_m2) \
+	const double *m1 = (src_m1); \
+	const double *m2 = (src_m2); \
+	for (int i = 0; i < OFFSET; i++, m1+=OFFSET, m2++) { \
+		dest_vu_m1[i][0].x[0] = m1[0]; \
+		dest_vu_m1[i][0].x[1] = m1[1]; \
+		dest_vu_m2[i][0].x[0] = m2[0*OFFSET]; \
+		dest_vu_m2[i][0].x[1] = m2[1*OFFSET]; \
+		dest_vu_m1[i][1].x[0] = m1[2]; \
+		dest_vu_m1[i][1].x[1] = m1[3]; \
+		dest_vu_m2[i][1].x[0] = m2[2*OFFSET]; \
+		dest_vu_m2[i][1].x[1] = m2[3*OFFSET]; \
+	}
+
 #define SSE_PREFETCH_MATRIX(src_m1, dest_vu_m1) \
 	const double *m1 = (src_m1); \
 	for (int i = 0; i < OFFSET; i++, m1++) { \
@@ -97,7 +126,7 @@ inline const char* getBeagleCPU4StateSSEName<double>(){ return "CPU-4State-SSE-D
 
 template<>
 inline const char* getBeagleCPU4StateSSEName<float>(){ return "CPU-4State-SSE-Single"; };
-    
+
 /*
  * Calculates partial likelihoods at a node when both children have states.
  */
@@ -279,7 +308,7 @@ void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartials(doubl
     	SSE_PREFETCH_MATRICES(matrices_q + w, matrices_r + w, vu_mq, vu_mr);
 
         for (int k = startPattern; k < endPattern; k++) {
-            
+
 #           if 1 && !defined(_WIN32)
             __builtin_prefetch (&partials_q[v+64]);
             __builtin_prefetch (&partials_r[v+64]);
@@ -368,6 +397,478 @@ void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartials(doubl
     }
 }
 
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPrePartialsPartials(double* destP,
+                                                                           const double*  partials_q,
+                                                                           const double*  matrices_q,
+                                                                           const double*  partials_r,
+                                                                           const double*  matrices_r,
+                                                                           int startPattern,
+                                                                           int endPattern) {
+
+    int patternDefficit = kPatternCount + kExtraPatterns - endPattern;
+
+    int v = 0;
+    int w = 0;
+
+    V_Real	destq_01, destq_23, destr_01, destr_23;
+    VecUnion vu_mq[OFFSET][2], vu_mr[OFFSET][2];
+    V_Real *destPvec = (V_Real *)destP;
+
+    for (int l = 0; l < kCategoryCount; l++) {
+        destPvec += startPattern*2;
+        v += startPattern*4;
+        /* Load transition-probability matrices into vectors */
+        SSE_PREFETCH_PRE_MATRICES(matrices_q + w, matrices_r + w, vu_mq, vu_mr);
+
+        for (int k = startPattern; k < endPattern; k++) {
+
+#           if 1 && !defined(_WIN32)
+            __builtin_prefetch (&partials_q[v+64]);
+            __builtin_prefetch (&partials_r[v+64]);
+//            __builtin_prefetch (destPvec+32,1,0);
+#           endif
+
+            V_Real vpr_0, vpr_1, vpr_2, vpr_3;
+            SSE_PREFETCH_PARTIALS(vpr_,partials_r,v);
+
+        	/* This would probably be faster on PPC/Altivec, which has a fused multiply-add
+               vector instruction */
+
+            destr_01 = VEC_MULT(vpr_0, vu_mr[0][0].vx);
+            destr_01 = VEC_MADD(vpr_1, vu_mr[1][0].vx, destr_01);
+            destr_01 = VEC_MADD(vpr_2, vu_mr[2][0].vx, destr_01);
+            destr_01 = VEC_MADD(vpr_3, vu_mr[3][0].vx, destr_01);
+            destr_23 = VEC_MULT(vpr_0, vu_mr[0][1].vx);
+            destr_23 = VEC_MADD(vpr_1, vu_mr[1][1].vx, destr_23);
+            destr_23 = VEC_MADD(vpr_2, vu_mr[2][1].vx, destr_23);
+            destr_23 = VEC_MADD(vpr_3, vu_mr[3][1].vx, destr_23);
+
+            V_Real vpq_0, vpq_1, vpq_2, vpq_3;
+            SSE_SCHUR_PRODUCT_PARTIALS(vpq_, partials_q, v, destr_);
+
+            destq_01 = VEC_MULT(vpq_0, vu_mq[0][0].vx);
+            destq_01 = VEC_MADD(vpq_1, vu_mq[1][0].vx, destq_01);
+            destq_01 = VEC_MADD(vpq_2, vu_mq[2][0].vx, destq_01);
+            destq_01 = VEC_MADD(vpq_3, vu_mq[3][0].vx, destq_01);
+            destq_23 = VEC_MULT(vpq_0, vu_mq[0][1].vx);
+            destq_23 = VEC_MADD(vpq_1, vu_mq[1][1].vx, destq_23);
+            destq_23 = VEC_MADD(vpq_2, vu_mq[2][1].vx, destq_23);
+            destq_23 = VEC_MADD(vpq_3, vu_mq[3][1].vx, destq_23);
+
+            destPvec[0] = destq_01;
+            destPvec[1] = destq_23;
+            destPvec += 2;
+
+                /* VEC_STORE did demonstrate a measurable performance gain as
+               it copies all (2/4) values to memory simultaneously;
+               I can no longer reproduce the performance gain (?) */
+
+            v += 4;
+        }
+        w += OFFSET*4;
+        if (kExtraPatterns) {
+            destPvec += kExtraPatterns * 2;
+            v += kExtraPatterns * 4;
+        }
+        destPvec += patternDefficit * 2;
+        v += patternDefficit * 4;
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPrePartialsStates(double* destP,
+                                                                              const double*  partials_q,
+                                                                              const double*  matrices_q,
+                                                                              const int*     states_r,
+                                                                              const double*  matrices_r,
+                                                                              int startPattern,
+                                                                              int endPattern) {
+
+    int patternDefficit = kPatternCount + kExtraPatterns - endPattern;
+
+    int v = 0;
+    int w = 0;
+
+    V_Real	destq_01, destq_23, destr_01, destr_23;
+    VecUnion vu_mq[OFFSET][2], vu_mr[OFFSET][2];
+    V_Real *destPvec = (V_Real *)destP;
+
+    for (int l = 0; l < kCategoryCount; l++) {
+        destPvec += startPattern*2;
+        v += startPattern*4;
+        /* Load transition-probability matrices into vectors */
+        SSE_PREFETCH_PRE_MATRICES(matrices_q + w, matrices_r + w, vu_mq, vu_mr);
+
+        for (int k = startPattern; k < endPattern; k++) {
+
+            /* This would probably be faster on PPC/Altivec, which has a fused multiply-add
+               vector instruction */
+
+            const int state_r = states_r[k];
+
+            V_Real destr_01, destr_23;
+            destr_01 = vu_mr[state_r][0].vx;
+            destr_23 = vu_mr[state_r][1].vx;
+
+            V_Real vpq_0, vpq_1, vpq_2, vpq_3;
+            SSE_SCHUR_PRODUCT_PARTIALS(vpq_, partials_q, v, destr_);
+
+
+            destq_01 = VEC_MULT(vpq_0, vu_mq[0][0].vx);
+            destq_01 = VEC_MADD(vpq_1, vu_mq[1][0].vx, destq_01);
+            destq_01 = VEC_MADD(vpq_2, vu_mq[2][0].vx, destq_01);
+            destq_01 = VEC_MADD(vpq_3, vu_mq[3][0].vx, destq_01);
+            destq_23 = VEC_MULT(vpq_0, vu_mq[0][1].vx);
+            destq_23 = VEC_MADD(vpq_1, vu_mq[1][1].vx, destq_23);
+            destq_23 = VEC_MADD(vpq_2, vu_mq[2][1].vx, destq_23);
+            destq_23 = VEC_MADD(vpq_3, vu_mq[3][1].vx, destq_23);
+
+
+            destPvec[0] = destq_01;
+            destPvec[1] = destq_23;
+            destPvec += 2;
+
+            /* VEC_STORE did demonstrate a measurable performance gain as
+           it copies all (2/4) values to memory simultaneously;
+           I can no longer reproduce the performance gain (?) */
+
+            v += 4;
+        }
+        w += OFFSET*4;
+        if (kExtraPatterns) {
+            destPvec += kExtraPatterns * 2;
+            v += kExtraPatterns * 4;
+        }
+        destPvec += patternDefficit * 2;
+        v += patternDefficit * 4;
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcCrossProductsPartials(const double *postOrderPartial,
+                                                                                const double *preOrderPartial,
+                                                                                const double *categoryRates,
+                                                                                const double *categoryWeights,
+                                                                                const double edgeLength,
+                                                                                double *outCrossProducts,
+                                                                                double *outSumSquaredDerivatives) {
+#if 0
+    return BeagleCPU4StateImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcCrossProductsPartials(postOrderPartial, preOrderPartial,
+                                                                                   categoryRates, categoryWeights,
+                                                                                   edgeLength, outCrossProducts,
+                                                                                   outSumSquaredDerivatives);
+#else
+
+    std::array<V_Real, 8> vAcrossPatterns;
+    vAcrossPatterns.fill(V_Real());
+
+    std::array<V_Real, 8> vWithinPattern;
+
+    for (int pattern = 0; pattern < kPatternCount; pattern++) {
+
+        vWithinPattern.fill(V_Real());
+
+        V_Real patternDenominator = VEC_SETZERO();
+
+        for (int category = 0; category < kCategoryCount; category++) {
+
+            const V_Real scale = VEC_SPLAT(categoryRates[category] * edgeLength);
+            const V_Real weight = VEC_SPLAT(categoryWeights[category]);
+
+            const int patternIndex = category * kPatternCount + pattern; // Bad memory access
+            const int v = patternIndex * 4;
+
+            V_Real pre0, pre1, pre2, pre3;
+            SSE_PREFETCH_PARTIALS(pre, preOrderPartial, v);
+
+            V_Real post01, post23;
+            SSE_PREFETCH_VECTORIZED_PARTIALS(post, postOrderPartial, v);
+
+            V_Real denominator = SSE_VECTORIZED_INNER_PRODUCT(tmp_pre, post);
+            patternDenominator = VEC_MADD(denominator, weight, patternDenominator);
+
+            V_Real weightScale = VEC_MULT(weight, scale);
+            post01 = VEC_MULT(post01, weightScale);
+            post23 = VEC_MULT(post23, weightScale);
+
+            vWithinPattern[0 * 2 + 0] = VEC_MADD(pre0, post01, vWithinPattern[0 * 2 + 0]);
+            vWithinPattern[0 * 2 + 1] = VEC_MADD(pre0, post23, vWithinPattern[0 * 2 + 1]);
+
+            vWithinPattern[1 * 2 + 0] = VEC_MADD(pre1, post01, vWithinPattern[1 * 2 + 0]);
+            vWithinPattern[1 * 2 + 1] = VEC_MADD(pre1, post23, vWithinPattern[1 * 2 + 1]);
+
+            vWithinPattern[2 * 2 + 0] = VEC_MADD(pre2, post01, vWithinPattern[2 * 2 + 0]);
+            vWithinPattern[2 * 2 + 1] = VEC_MADD(pre2, post23, vWithinPattern[2 * 2 + 1]);
+
+            vWithinPattern[3 * 2 + 0] = VEC_MADD(pre3, post01, vWithinPattern[3 * 2 + 0]);
+            vWithinPattern[3 * 2 + 1] = VEC_MADD(pre3, post23, vWithinPattern[3 * 2 + 1]);
+        }
+
+        const V_Real patternWeight = VEC_DIV(VEC_SPLAT(gPatternWeights[pattern]), patternDenominator);
+        for (int k = 0; k < 8; k++) {
+            vAcrossPatterns[k] = VEC_MADD(vWithinPattern[k], patternWeight, vAcrossPatterns[k]);
+        }
+    }
+
+    for (int k = 0; k < 8; k++) {
+        VEC_STORE(&outCrossProducts[k * 2],
+                VEC_ADD(VEC_LOAD(&outCrossProducts[k * 2]), vAcrossPatterns[k]));
+    }
+#endif
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcCrossProductsStates(const int *tipStates,
+                                                                              const double *preOrderPartial,
+                                                                              const double *categoryRates,
+                                                                              const double *categoryWeights,
+                                                                              const double edgeLength,
+                                                                              double *outCrossProducts,
+                                                                              double *outSumSquaredDerivatives) {
+
+    return BeagleCPU4StateImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcCrossProductsStates(tipStates, preOrderPartial, categoryRates,
+                                                                                 categoryWeights, edgeLength, outCrossProducts,
+                                                                                 outSumSquaredDerivatives);
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE template <bool DoDerivatives, bool DoSum, bool DoSumSquared>
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::accumulateDerivativesImpl(
+        double* outDerivatives,
+        double* outSumDerivatives,
+        double* outSumSquaredDerivatives) {
+
+    V_Real vSum = VEC_SETZERO();
+    V_Real vSumSquared = VEC_SETZERO();
+
+    int k = 0;
+    for (; k < kPatternCount - 1; k += 2) {
+
+        V_Real numerator = VEC_LOAD(grandNumeratorDerivTmp + k);
+        V_Real denominator = VEC_LOAD(grandDenominatorDerivTmp + k);
+        V_Real derivative = VEC_DIV(numerator, denominator);
+        V_Real patternWeight = VEC_LOAD(gPatternWeights + k);
+
+        if (DoDerivatives) {
+            VEC_STOREU(outDerivatives + k, derivative);
+        }
+
+        if (DoSum) {
+            vSum = VEC_MADD(derivative, patternWeight, vSum);
+        }
+
+        if (DoSumSquared) {
+            derivative = VEC_MULT(derivative, derivative);
+            vSumSquared = VEC_MADD(derivative, patternWeight, vSumSquared);
+        }
+    }
+
+    double sum;
+    double sumSquared;
+
+    if (DoSum) {
+        sum = _mm_cvtsd_f64(VEC_ADD(vSum, VEC_SWAP(vSum)));
+    }
+
+    if (DoSumSquared) {
+        sumSquared = _mm_cvtsd_f64(VEC_ADD(vSumSquared, VEC_SWAP(vSumSquared)));
+    }
+
+    for (; k < kPatternCount; ++k) {
+        double derivative = grandNumeratorDerivTmp[k] / grandDenominatorDerivTmp[k];
+        if (DoDerivatives) {
+            outDerivatives[k] = derivative;
+        }
+        if (DoSum) {
+            sum += derivative * gPatternWeights[k];
+        }
+        if (DoSumSquared) {
+            sumSquared += derivative * derivative * gPatternWeights[k];
+        }
+    }
+
+    if (DoSum) {
+        *outSumDerivatives = sum;
+    }
+
+    if (DoSumSquared) {
+        *outSumSquaredDerivatives = sumSquared;
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE template <bool DoDerivatives, bool DoSum>
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::accumulateDerivativesDispatch2(
+        double* outDerivatives,
+        double* outSumDerivatives,
+        double* outSumSquaredDerivatives) {
+
+    if (outSumSquaredDerivatives == NULL) {
+        accumulateDerivativesImpl<DoDerivatives, DoSum, false>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    } else {
+        accumulateDerivativesImpl<DoDerivatives, DoSum, true>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE template <bool DoDerivatives>
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::accumulateDerivativesDispatch1(
+        double* outDerivatives,
+        double* outSumDerivatives,
+        double* outSumSquaredDerivatives) {
+
+    if (outSumDerivatives == NULL) {
+        accumulateDerivativesDispatch2<DoDerivatives, false>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    } else {
+        accumulateDerivativesDispatch2<DoDerivatives, true>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    }
+}
+
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::accumulateDerivatives(double* outDerivatives,
+                                                                            double* outSumDerivatives,
+                                                                            double* outSumSquaredDerivatives) {
+    if (outDerivatives == NULL) {
+        accumulateDerivativesDispatch1<false>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    } else {
+        accumulateDerivativesDispatch1<true>(
+                outDerivatives, outSumDerivatives, outSumSquaredDerivatives);
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcEdgeLogDerivativesPartials(const double* postOrderPartial,
+                                                                                     const double* preOrderPartial,
+                                                                                     const int firstDerivativeIndex,
+                                                                                     const int secondDerivativeIndex,
+                                                                                     const double* categoryRates,
+                                                                                     const double* categoryWeights,
+                                                                                     const int scalingFactorsIndex,
+                                                                                     double* outDerivatives,
+                                                                                     double* outSumDerivatives,
+                                                                                     double* outSumSquaredDerivatives) {
+    double* cl_p = integrationTmp;
+    memset(cl_p, 0, (kPatternCount * kStateCount)*sizeof(double));
+
+    const double* cl_r = preOrderPartial;
+    const double* wt = categoryWeights;
+
+    V_Real * vcl_r = (V_Real *)cl_r;
+
+    int v = 0;
+    int w = 0;
+    const double* transMatrix = gTransitionMatrices[firstDerivativeIndex];
+
+    for (int l = 0; l < kCategoryCount; l++) {
+
+        /* Load transition-probability matrix into vectors */
+        VecUnion vu_m[OFFSET][2];
+        SSE_PREFETCH_MATRIX(transMatrix + w, vu_m);
+
+        V_Real * vcl_p = (V_Real *)cl_p;
+        V_Real vwt = VEC_SPLAT(wt[l]);
+        for (int k = 0; k < kPatternCount; k++) {
+
+            /* This would probably be faster on PPC/Altivec, which has a fused multiply-add
+               vector instruction */
+
+            V_Real vcl_q0, vcl_q1, vcl_q2, vcl_q3;
+            SSE_PREFETCH_PARTIALS(vcl_q,postOrderPartial,v);
+
+            V_Real vclp_01, vclp_23;
+            vclp_01 = VEC_MULT(vcl_q0, vu_m[0][0].vx);
+            vclp_01 = VEC_MADD(vcl_q1, vu_m[1][0].vx, vclp_01);
+            vclp_01 = VEC_MADD(vcl_q2, vu_m[2][0].vx, vclp_01);
+            vclp_01 = VEC_MADD(vcl_q3, vu_m[3][0].vx, vclp_01);
+            vclp_23 = VEC_MULT(vcl_q0, vu_m[0][1].vx);
+            vclp_23 = VEC_MADD(vcl_q1, vu_m[1][1].vx, vclp_23);
+            vclp_23 = VEC_MADD(vcl_q2, vu_m[2][1].vx, vclp_23);
+            vclp_23 = VEC_MADD(vcl_q3, vu_m[3][1].vx, vclp_23);
+
+            vclp_01 = VEC_MULT(vclp_01, *vcl_r);
+            tmp_vcl_q01 = VEC_MULT(tmp_vcl_q01, *vcl_r++);
+            vclp_23 = VEC_MULT(vclp_23, *vcl_r);
+            tmp_vcl_q23 = VEC_MULT(tmp_vcl_q23, *vcl_r++);
+
+            V_Real vnumer = VEC_ADD(vclp_01, vclp_23);
+            vnumer = VEC_ADD(vnumer, VEC_SWAP(vnumer));
+
+            V_Real vdenom = VEC_ADD(tmp_vcl_q01, tmp_vcl_q23);
+            vdenom = VEC_ADD(vdenom, VEC_SWAP(vdenom));
+
+            double numer = _mm_cvtsd_f64(vnumer) * wt[l];
+            double denon = _mm_cvtsd_f64(vdenom) * wt[l];
+
+            grandNumeratorDerivTmp[k] += numer; // TODO Merge [numer, denom] into single SSE transactions
+            grandDenominatorDerivTmp[k] += denon;
+
+            v += 4;
+        }
+        w += 4*OFFSET;
+        if (kExtraPatterns) {
+            vcl_r += 2 * kExtraPatterns;
+            v += 4 * kExtraPatterns;
+        }
+    }
+}
+
+BEAGLE_CPU_4_SSE_TEMPLATE
+void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcEdgeLogDerivativesStates(const int* tipStates,
+                                                                                   const double* preOrderPartial,
+                                                                                   const int firstDerivativeIndex,
+                                                                                   const int secondDerivativeIndex,
+                                                                                   const double* categoryRates,
+                                                                                   const double* categoryWeights,
+                                                                                   double *outDerivatives,
+                                                                                   double *outSumDerivatives,
+                                                                                   double *outSumSquaredDerivatives) {
+    double* cl_p = integrationTmp;
+    memset(cl_p, 0, (kPatternCount * kStateCount)*sizeof(double));
+
+    const double* cl_r = preOrderPartial;
+    const double* wt = categoryWeights;
+
+    const int* statesChild = tipStates;
+
+    int w = 0;
+    V_Real *vcl_r = (V_Real *)cl_r;
+    const double* transMatrix = gTransitionMatrices[firstDerivativeIndex];
+
+    for (int l = 0; l < kCategoryCount; l++) {
+
+        /* Load transition-probability matrix into vectors */
+        VecUnion vu_m[OFFSET][2];
+        SSE_PREFETCH_MATRIX(transMatrix + w, vu_m);
+
+        V_Real * vcl_p = (V_Real *)cl_p;
+        V_Real vwt = VEC_SPLAT(wt[l]);
+        for (int k = 0; k < kPatternCount; k++) {
+
+            const int stateChild = statesChild[k];
+
+            V_Real p01, p23;
+            p01 = VEC_MULT(vu_m[stateChild][0].vx, *vcl_r++);
+            p23 = VEC_MULT(vu_m[stateChild][1].vx, *vcl_r++);
+
+            V_Real vnumer = VEC_ADD(p01, p23);
+            vnumer = VEC_ADD(vnumer, VEC_SWAP(vnumer));
+
+            double numer = _mm_cvtsd_f64(vnumer);
+            double denom = cl_r[stateChild & 3]; cl_r += 4;
+
+            grandNumeratorDerivTmp[k] += numer * wt[l];
+            grandDenominatorDerivTmp[k] += denom * wt[l];
+        }
+        w += OFFSET*4;
+        vcl_r += 2 * kExtraPatterns;
+        cl_r += 4 * kExtraPatterns;
+    }
+}
+
 BEAGLE_CPU_4_SSE_TEMPLATE
 void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartialsFixedScaling(double* destP,
 		                                                                                   const double* partials_q,
@@ -400,7 +901,7 @@ void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartialsFixedS
             __builtin_prefetch (&partials_r[v+64]);
             //            __builtin_prefetch (destPvec+32,1,0);
 #           endif
-            
+
             // Prefetch scale factor
 //            const V_Real scaleFactor = VEC_LOAD_SCALAR(scaleFactors + k);
         	// Option below appears faster, why?
@@ -448,7 +949,7 @@ void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartialsFixedS
     }
 }
 
-    
+
 BEAGLE_CPU_4_SSE_TEMPLATE
 void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_FLOAT>::calcPartialsPartialsAutoScaling(float* destP,
                                                          const float*  partials_q,
@@ -479,7 +980,7 @@ void BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcPartialsPartialsAutoSc
                                                                 matrices_r,
                                                                 activateScaling);
 }
-    
+
 BEAGLE_CPU_4_SSE_TEMPLATE
 int BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_FLOAT>::calcEdgeLogLikelihoods(const int parIndex,
                                                           const int childIndex,
@@ -622,7 +1123,7 @@ int BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::calcEdgeLogLikelihoods(cons
 
     if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
-        
+
     return returnCode;
 }
 
@@ -802,7 +1303,7 @@ int BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_FLOAT>::getPaddedPatternsModulus() {
 //	return 4;  // For single-precision, can operate on 4 patterns at a time
 	// TODO Vectorize final log operations over patterns
 }
-    
+
 BEAGLE_CPU_4_SSE_TEMPLATE
 int BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::getPaddedPatternsModulus() {
 //	return 2;  // For double-precision, can operate on 2 patterns at a time
@@ -819,7 +1320,7 @@ const char* BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_DOUBLE>::getName() {
     return  getBeagleCPU4StateSSEName<double>();
 }
 
-    
+
 BEAGLE_CPU_4_SSE_TEMPLATE
 const long BeagleCPU4StateSSEImpl<BEAGLE_CPU_4_SSE_FLOAT>::getFlags() {
 	return  BEAGLE_FLAG_COMPUTATION_SYNCH |
@@ -907,6 +1408,7 @@ const long BeagleCPU4StateSSEImplFactory<double>::getFlags() {
            BEAGLE_FLAG_SCALERS_LOG | BEAGLE_FLAG_SCALERS_RAW |
            BEAGLE_FLAG_EIGEN_COMPLEX | BEAGLE_FLAG_EIGEN_REAL|
            BEAGLE_FLAG_INVEVEC_STANDARD | BEAGLE_FLAG_INVEVEC_TRANSPOSED |
+           BEAGLE_FLAG_PREORDER_TRANSPOSE_MANUAL | BEAGLE_FLAG_PREORDER_TRANSPOSE_AUTO |
            BEAGLE_FLAG_FRAMEWORK_CPU;
 }
 
@@ -921,6 +1423,7 @@ const long BeagleCPU4StateSSEImplFactory<float>::getFlags() {
            BEAGLE_FLAG_SCALERS_LOG | BEAGLE_FLAG_SCALERS_RAW |
            BEAGLE_FLAG_EIGEN_COMPLEX | BEAGLE_FLAG_EIGEN_REAL |
            BEAGLE_FLAG_INVEVEC_STANDARD | BEAGLE_FLAG_INVEVEC_TRANSPOSED |
+           BEAGLE_FLAG_PREORDER_TRANSPOSE_MANUAL | BEAGLE_FLAG_PREORDER_TRANSPOSE_AUTO |
            BEAGLE_FLAG_FRAMEWORK_CPU;
 }
 
