@@ -6,19 +6,9 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @author Andrew Rambaut
  * @author Marc Suchard
@@ -50,7 +40,6 @@
 #define BEAGLE_CPU_FACTORY_GENERIC	REALTYPE
 #define BEAGLE_CPU_FACTORY_TEMPLATE	template <typename REALTYPE>
 
-
 #define T_PAD_DEFAULT   1   // Pad transition matrix rows with an extra 1.0 for ambiguous characters
 #define P_PAD_DEFAULT   0   // No partials padding necessary for non-SSE implementations
 
@@ -84,8 +73,8 @@ protected:
 
     int kPartialsSize;  /// stored for convenience. kPartialsSize = kStateCount*kPatternCount
     int kMatrixSize; /// stored for convenience. kMatrixSize = kStateCount*(kStateCount + 1)
-    
-    int kInternalPartialsBufferCount; 
+
+    int kInternalPartialsBufferCount;
 
     int kPartitionCount;
     int kMaxPartitionCount;
@@ -94,7 +83,7 @@ protected:
     int kMinPatternCount;
 
     long kFlags;
-    
+
     REALTYPE realtypeMin;
     int scalingExponentThreshold;
 
@@ -106,19 +95,19 @@ protected:
     int* gPatternPartitions;
     int* gPatternPartitionsStartPatterns;
     int* gPatternsNewOrder;
-    
+
     REALTYPE** gCategoryWeights;
     REALTYPE** gStateFrequencies;
-    
+
     //@ the size of these pointers are known at alloc-time, so the partials and
     //      tipStates field should be switched to vectors of vectors (to make
     //      memory management less error prone
     REALTYPE** gPartials;
     int** gTipStates;
     REALTYPE** gScaleBuffers;
-    
+
     signed short** gAutoScaleBuffers;
-    
+
     int* gActiveScalingFactors;
 
     // There will be kMatrixCount transitionMatrices.
@@ -129,7 +118,14 @@ protected:
     REALTYPE* integrationTmp;
     REALTYPE* firstDerivTmp;
     REALTYPE* secondDerivTmp;
-    
+
+//    REALTYPE* cLikelihoodTmp;
+    REALTYPE* grandDenominatorDerivTmp;
+    REALTYPE* grandNumeratorDerivTmp;
+    REALTYPE* crossProductNumeratorTmp;
+//    REALTYPE* grandNumeratorLowerBoundDerivTmp;
+//    REALTYPE* grandNumeratorUpperBoundDerivTmp;
+
     REALTYPE* outLogLikelihoodsTmp;
     REALTYPE* outFirstDerivativesTmp;
     REALTYPE* outSecondDerivativesTmp;
@@ -196,6 +192,14 @@ public:
     int setTipPartials(int tipIndex,
                        const double* inPartials);
 
+    // set the pre-order partials for root node
+    //
+    // bufferIndices the indices of root nodes (may countain multiple)
+    // stateFrequenciesIndices the indices of state frequencies at root
+    // count number of root nodes
+    int setRootPrePartials(const int* bufferIndices,
+                           const int* stateFrequenciesIndices,
+                           int count);
 
     int setPartials(int bufferIndex,
                     const double* inPartials);
@@ -216,16 +220,16 @@ public:
                               const double* inEigenValues);
 
     int setStateFrequencies(int stateFrequenciesIndex,
-                            const double* inStateFrequencies);    
-    
+                            const double* inStateFrequencies);
+
     int setCategoryWeights(int categoryWeightsIndex,
                            const double* inCategoryWeights);
-    
-    int setPatternWeights(const double* inPatternWeights); 
+
+    int setPatternWeights(const double* inPatternWeights);
 
     int setPatternPartitions(int partitionCount,
                              const int* inPatternPartitions);
-    
+
     // set the vector of category rates
     //
     // categoryRates an array containing categoryCount rate scalers
@@ -237,7 +241,10 @@ public:
     int setTransitionMatrix(int matrixIndex,
                             const double* inMatrix,
                             double paddedValue);
-    
+
+    int setDifferentialMatrix(int matrixIndex,
+                              const double* inMatrix);
+
     int setTransitionMatrices(const int* matrixIndices,
                               const double* inMatrices,
                               const double* paddedValues,
@@ -246,14 +253,19 @@ public:
     int getTransitionMatrix(int matrixIndex,
     						double* outMatrix);
 
-    ///////////////////////////
-    //---TODO: Epoch model---//
-    ///////////////////////////
+    int convolveTransitionMatrices(const int* firstIndices,
+                                   const int* secondIndices,
+                                   const int* resultIndices,
+                                   int count);
 
-	int convolveTransitionMatrices( const int* firstIndices,
-            const int* secondIndices,
-            const int* resultIndices,
-            int count);
+	int addTransitionMatrices(const int* firstIndices,
+	                          const int* secondIndices,
+	                          const int* resultIndices,
+	                          int matrixCount);
+
+    int transposeTransitionMatrices(const int* inputIndices,
+                                    const int* resultIndices,
+                                    int matrixCount);
 
     // calculate a transition probability matrices for a given list of node. This will
     // calculate for all categories (and all matrices if more than one is being used).
@@ -268,6 +280,14 @@ public:
                                  const double* edgeLengths,
                                  int count);
 
+    int updateTransitionMatricesWithModelCategories(int* eigenIndices,
+                                 const int* probabilityIndices,
+                                 const int* firstDerivativeIndices,
+                                 const int* secondDerivativeIndices,
+                                 const double* edgeLengths,
+                                 int count);
+
+
     int updateTransitionMatricesWithMultipleModels(const int* eigenIndices,
                                                    const int* categoryRateIndices,
                                                    const int* probabilityIndices,
@@ -286,9 +306,16 @@ public:
                        int operationCount,
                        int cumulativeScalingIndex);
 
+    int updatePrePartials(const int *operations,
+                          int operationCount,
+                          int cumulativeScalingIndex);
+
     int updatePartialsByPartition(const int* operations,
                                   int operationCount);
 
+    int updatePrePartialsByPartition(const int* operations,
+                                     int operationCount);
+
     // Block until all calculations that write to the specified partials have completed.
     //
     // This function is optional and only has to be called by clients that "recycle" partials.
@@ -329,11 +356,11 @@ public:
     int resetScaleFactorsByPartition(int cumulativeScalingIndex, int partitionIndex);
 
     int copyScaleFactors(int destScalingIndex,
-                         int srcScalingIndex);    
-                         
+                         int srcScalingIndex);
+
 	int getScaleFactors(int srcScalingIndex,
-                        double* scaleFactors);                          
-    
+                        double* scaleFactors);
+
     // calculate the site log likelihoods at a particular node
     //
     // rootNodeIndex the index of the root
@@ -387,14 +414,34 @@ public:
                                                double* outSumFirstDerivative,
                                                double* outSumSecondDerivativeByPartition,
                                                double* outSumSecondDerivative);
-    
+
+    int calculateEdgeDerivatives(const int *postBufferIndices,
+                                 const int *preBufferIndices,
+                                 const int *derivativeMatrixIndices,
+                                 const int *categoryWeightsIndices,
+                                 const int *categoryRatesIndices,
+                                 const int *cumulativeScaleIndices,
+                                 int count,
+                                 double *outDerivatives,
+                                 double *outSumDerivatives,
+                                 double *outSumSquaredDerivatives);
+
+    int calculateCrossProducts(const int *postBufferIndices,
+                               const int *preBufferIndices,
+                               const int *categoryRatesIndices,
+                               const int *categoryWeightsIndices,
+                               const double *edgeLengths,
+                               int count,
+                               double *outSumDerivatives,
+                               double *outSumSquaredDerivatives);
+
     int getLogLikelihood(double* outSumLogLikelihood);
 
     int getDerivatives(double* outSumFirstDerivative,
                        double* outSumSecondDerivative);
 
     int getSiteLogLikelihoods(double* outLogLikelihoods);
-    
+
     int getSiteDerivatives(double* outFirstDerivatives,
                            double* outSecondDerivatives);
 
@@ -410,6 +457,77 @@ protected:
                            int operationCount,
                            int cumulativeScalingIndex);
 
+    virtual int upPrePartials(bool byPartition,
+                              const int* operations,
+                              int count,
+                              int cumulativeScaleIndex);
+
+    virtual int calcEdgeLogDerivatives(const int *postBufferIndices,
+                                       const int *preBufferIndices,
+                                       const int *firstDerivativeIndices,
+                                       const int *secondDerivativeIndices,
+                                       const int *categoryWeightsIndices,
+                                       const int *categoryRatesIndices,
+                                       const int *cumulativeScaleIndices,
+                                       int count,
+                                       double *siteLogLikelihoods,
+                                       double *outLogFirstDerivatives,
+                                       double *outLogDiagonalSecondDerivatives);
+
+    virtual void calcEdgeLogDerivativesStates(const int *tipStates,
+                                              const REALTYPE *preOrderPartial,
+                                              const int firstDerivativeIndex,
+                                              const int secondDerivativeIndex,
+                                              const double *categoryRates,
+                                              const REALTYPE *categoryWeights,
+//                                              const REALTYPE *cumulativeScaleBuffer,
+                                              double *siteLogLikelihoods,
+                                              double *outLogFirstDerivatives,
+                                              double *outLogDiagonalSecondDerivatives);
+
+    virtual void calcEdgeLogDerivativesPartials(const REALTYPE *postOrderPartial,
+                                                const REALTYPE *preOrderPartial,
+                                                const int firstDerivativeIndex,
+                                                const int secondDerivativeIndex,
+                                                const double *categoryRates,
+                                                const REALTYPE *categoryWeights,
+                                                const int scalingFactorsIndex,
+//                                                const REALTYPE *cumulativeScaleBuffer,
+                                                double *siteLogLikelihoods,
+                                                double *outLogFirstDerivatives,
+                                                double *outLogDiagonalSecondDerivatives);
+
+    virtual int calcCrossProducts(const int *postBufferIndices,
+                                  const int *preBufferIndices,
+                                  const int *categoryRatesIndices,
+                                  const int *categoryWeightsIndices,
+                                  const double *edgeLengths,
+                                  int count,
+                                  double *outSumDerivatives,
+                                  double *outSumSquaredDerivatives);
+
+    virtual void calcCrossProductsStates(const int *tipStates,
+                                         const REALTYPE *preOrderPartial,
+                                         const double *categoryRates,
+                                         const REALTYPE *categoryWeights,
+                                         const double edgeLength,
+                                         double *outCrossProducts,
+                                         double *outSumSquaredDerivatives);
+
+    virtual void calcCrossProductsPartials(const REALTYPE *postOrderPartial,
+                                           const REALTYPE *preOrderPartial,
+                                           const double *categoryRates,
+                                           const REALTYPE *categoryWeights,
+                                           const double edgeLength,
+                                           double *outCrossProducts,
+                                           double *outSumSquaredDerivatives);
+
+    virtual void resetDerivativeTemporaries();
+
+    virtual void accumulateDerivatives(double* outDerivatives,
+                               double* outSumDerivatives,
+                               double* outSumSquaredDerivatives);
+
     virtual void autoPartitionPartialsOperations(const int* operations,
                                                  int* partitionOperations,
                                                  int count,
@@ -445,12 +563,33 @@ protected:
                                       int startPattern,
                                       int endPattern);
 
+    virtual void calcPrePartialsPartials(REALTYPE* destP,
+                                         const REALTYPE* partials1,
+                                         const REALTYPE* matrices1,
+                                         const REALTYPE* partials2,
+                                         const REALTYPE* matrices2,
+                                         int startPattern,
+                                         int endPattern);
+
+    virtual void calcPrePartialsStates(REALTYPE* destP,
+                                         const REALTYPE* partials1,
+                                         const REALTYPE* matrices1,
+                                         const int* states2,
+                                         const REALTYPE* matrices2,
+                                         int startPattern,
+                                         int endPattern);
+
     virtual int calcRootLogLikelihoods(const int bufferIndex,
                                         const int categoryWeightsIndex,
                                         const int stateFrequenciesIndex,
                                         const int scaleBufferIndex,
                                         double* outSumLogLikelihood);
 
+    virtual int calcRootLogLikelihoodsPerCategory(const int bufferIndex,
+                                                  const int stateFrequenciesIndex,
+                                                  const int scaleBufferIndex,
+                                                  double* outSumLogLikelihood);
+
     virtual void calcRootLogLikelihoodsByPartitionAsync(const int* bufferIndices,
                                                        const int* categoryWeightsIndices,
                                                        const int* stateFrequenciesIndices,
@@ -473,14 +612,14 @@ protected:
                                                   const int* partitionIndices,
                                                   int partitionCount,
                                                   double* outSumLogLikelihoodByPartition);
-    
+
     virtual int calcRootLogLikelihoodsMulti(const int* bufferIndices,
                                              const int* categoryWeightsIndices,
                                              const int* stateFrequenciesIndices,
                                              const int* scaleBufferIndices,
                                              int count,
                                              double* outSumLogLikelihood);
-    
+
     virtual int calcEdgeLogLikelihoods(const int parentBufferIndex,
                                         const int childBufferIndex,
                                         const int probabilityIndex,
@@ -541,7 +680,7 @@ protected:
                                             const int* scalingFactorsIndices,
                                             int count,
                                             double* outSumLogLikelihood);
-    
+
     virtual int calcEdgeLogLikelihoodsFirstDeriv(const int parentBufferIndex,
                                                   const int childBufferIndex,
                                                   const int probabilityIndex,
@@ -551,7 +690,7 @@ protected:
                                                   const int scalingFactorsIndex,
                                                   double* outSumLogLikelihood,
                                                   double* outSumFirstDerivative);
-	
+
     virtual int calcEdgeLogLikelihoodsSecondDeriv(const int parentBufferIndex,
                                                    const int childBufferIndex,
                                                    const int probabilityIndex,
@@ -582,6 +721,15 @@ protected:
                                                 int startPattern,
                                                 int endPattern);
 
+    virtual void calcPreStatesPartialsFixedScaling(REALTYPE* destP,
+                                                   const int* states1,
+                                                   const REALTYPE* matrices1,
+                                                   const REALTYPE* partials2,
+                                                   const REALTYPE* matrices2,
+                                                   const REALTYPE* scaleFactors,
+                                                   int startPattern,
+                                                   int endPattern);
+
     virtual void calcPartialsPartialsFixedScaling(REALTYPE *destP,
                                             const REALTYPE *child0States,
                                             const REALTYPE *child0TransMat,
@@ -590,7 +738,7 @@ protected:
                                             const REALTYPE *scaleFactors,
                                             int startPattern,
                                             int endPattern);
-    
+
     virtual void calcPartialsPartialsAutoScaling(REALTYPE* destP,
                                                   const REALTYPE* partials1,
                                                   const REALTYPE* matrices1,
@@ -608,7 +756,7 @@ protected:
                                             REALTYPE *cumulativeScaleFactors,
                                             const int fillWithOnes,
                                             const int partitionIndex);
-    
+
     virtual void autoRescalePartials(REALTYPE *destP,
     		                     signed short *scaleFactors);
 
@@ -618,6 +766,22 @@ protected:
 
     void threadWaiting(threadData* tData);
 
+private:
+
+    template <bool DoDerivatives>
+    void accumulateDerivativesDispatch1(double* outDerivatives,
+                               double* outSumDerivatives,
+                               double* outSumSquaredDerivatives);
+
+    template <bool DoDerivatives, bool DoSum>
+    void accumulateDerivativesDispatch2(double* outDerivatives,
+                               double* outSumDerivatives,
+                               double* outSumSquaredDerivatives);
+
+    template <bool DoDerivatives, bool DoSum, bool DoSumSquared>
+    void accumulateDerivativesImpl(double* outDerivatives,
+                               double* outSumDerivatives,
+                               double* outSumSquaredDerivatives);
 };
 
 BEAGLE_CPU_FACTORY_TEMPLATE
