@@ -5,21 +5,10 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
- *
- * 
  * @brief GPU implementation header
  *
  * @author Marc Suchard
@@ -34,6 +23,8 @@
 #include "libhmsbeagle/config.h"
 #endif
 
+#include <vector>
+
 #include "libhmsbeagle/BeagleImpl.h"
 #include "libhmsbeagle/GPU/GPUImplDefs.h"
 #include "libhmsbeagle/GPU/GPUInterface.h"
@@ -64,11 +55,11 @@ class BeagleGPUImpl : public BeagleImpl {
 private:
     GPUInterface* gpu;
     KernelLauncher* kernels;
-    
+
     int kInitialized;
-    
+
     long kFlags;
-    
+
     int kTipCount;
     int kPartialsBufferCount;
     int kCompactBufferCount;
@@ -77,17 +68,17 @@ private:
     int kEigenDecompCount;
     int kMatrixCount;
     int kCategoryCount;
- 
+
     int kTipPartialsBufferCount;
     int kInternalPartialsBufferCount;
     int kBufferCount;
     int kScaleBufferCount;
-    
+
     int kPaddedStateCount;
     int kPaddedPatternCount;    // total # of patterns with padding so that kPaddedPatternCount
                                 //   * kPaddedStateCount is a multiple of 16
     int kSumSitesBlockCount;
-    
+
     int kPartialsSize;
     int kMatrixSize;
     int kEigenValuesSize;
@@ -97,50 +88,58 @@ private:
     int kLastTipPartialsBufferIndex;
 
     int kResultPaddedPatterns;
-    
+
     GPUPtr dIntegrationTmp;
     GPUPtr dOutFirstDeriv;
     GPUPtr dOutSecondDeriv;
     GPUPtr dPartialsTmp;
     GPUPtr dFirstDerivTmp;
     GPUPtr dSecondDerivTmp;
-    
+
     GPUPtr dSumLogLikelihood;
     GPUPtr dSumFirstDeriv;
     GPUPtr dSumSecondDeriv;
-    
-    GPUPtr dPatternWeights;    
-	
+
+	GPUPtr dMultipleDerivatives;
+	GPUPtr dMultipleDerivativeSum;
+
+    GPUPtr dPatternWeights;
+
     GPUPtr dBranchLengths;
-    
+
     GPUPtr dDistanceQueue;
-    
+
     GPUPtr dPtrQueue;
-	
+
+	GPUPtr dDerivativeQueue;
+
     GPUPtr dMaxScalingFactors;
     GPUPtr dIndexMaxScalingFactors;
-    
+
     GPUPtr dAccumulatedScalingFactors;
-    
+
     GPUPtr* dEigenValues;
     GPUPtr* dEvec;
     GPUPtr* dIevc;
-    
+
     GPUPtr* dWeights;
-    GPUPtr* dFrequencies; 
+    GPUPtr* dFrequencies;
 
     GPUPtr* dScalingFactors;
-    
+
     GPUPtr* dStates;
-    
+
     GPUPtr* dPartials;
     GPUPtr* dMatrices;
-    
+
     GPUPtr* dCompactBuffers;
     GPUPtr* dTipPartialsBuffers;
-    
+
     bool kUsingMultiGrid;
     bool kDerivBuffersInitialised;
+	int kMultipleDerivativesLength; // TODO Change to length (max node count used)
+	bool kUsingAutoTranspose;
+
     int kNumPatternBlocks;
     int kSitesPerBlock;
     int kSitesPerIntegrateBlock;
@@ -186,33 +185,37 @@ private:
     int* hPatternsNewOrder;
     int* hGridOpIndices;
 
+    int kExtraMatrixCount;
+
     unsigned int* hPtrQueue;
-    
+
+	unsigned int* hDerivativeQueue;
+
     double** hCategoryRates; // Can keep in double-precision
 
     Real* hPatternWeightsCache;
-        
+
     Real* hDistanceQueue;
-    
+
     Real* hWeightsCache;
     Real* hFrequenciesCache;
     Real* hLogLikelihoodsCache;
     Real* hPartialsCache;
     int* hStatesCache;
     Real* hMatrixCache;
-    
+
     int* hRescalingTrigger;
     GPUPtr dRescalingTrigger;
-    
+
     GPUPtr* dScalingFactorsMaster;
-    
+
     int* hStreamIndices;
 
-public:    
+public:
     BeagleGPUImpl();
-    
+
     virtual ~BeagleGPUImpl();
-    
+
     int createInstance(int tipCount,
                        int partialsBufferCount,
                        int compactBufferCount,
@@ -226,7 +229,7 @@ public:
                        int pluginResourceNumber,
                        long preferenceFlags,
                        long requirementFlags);
-    
+
     int getInstanceDetails(BeagleInstanceDetails* retunInfo);
 
     int setCPUThreadCount(int threadCount);
@@ -236,56 +239,68 @@ public:
 
     int setTipPartials(int tipIndex,
                        const double* inPartials);
-    
+
+    int setRootPrePartials(const int* bufferIndices,
+                           const int* stateFrequenciesIndices,
+                           int count);
+
     int setPartials(int bufferIndex,
                     const double* inPartials);
-    
+
     int getPartials(int bufferIndex,
 				    int scaleIndex,
                     double* outPartials);
-        
+
     int setEigenDecomposition(int eigenIndex,
                               const double* inEigenVectors,
                               const double* inInverseEigenVectors,
                               const double* inEigenValues);
-    
+
     int setStateFrequencies(int stateFrequenciesIndex,
-                            const double* inStateFrequencies);    
-    
+                            const double* inStateFrequencies);
+
     int setCategoryWeights(int categoryWeightsIndex,
                            const double* inCategoryWeights);
-    
+
     int setPatternWeights(const double* inPatternWeights);
 
     int setPatternPartitions(int partitionCount,
                              const int* inPatternPartitions);
-    
+
     int setCategoryRates(const double* inCategoryRates);
 
     int setCategoryRatesWithIndex(int categoryRatesIndex,
                                   const double* inCategoryRates);
-    
+
     int setTransitionMatrix(int matrixIndex,
                             const double* inMatrix,
                             double paddedValue);
-    
+
+    int setDifferentialMatrix(int matrixIndex,
+                              const double* inMatrix);
+
     int setTransitionMatrices(const int* matrixIndices,
                               const double* inMatrices,
                               const double* paddedValues,
-                              int count);    
-    
+                              int count);
+
     int getTransitionMatrix(int matrixIndex,
                             double* outMatrix);
 
-    ///////////////////////////
-    //---TODO: Epoch model---//
-    ///////////////////////////
-
   	int convolveTransitionMatrices(const int* firstIndices,
                                    const int* secondIndices,
                                    const int* resultIndices,
                                    int matrixCount);
 
+	int addTransitionMatrices(const int* firstIndices,
+							  const int* secondIndices,
+							  const int* resultIndices,
+							  int matrixCount);
+
+    int transposeTransitionMatrices(const int* inputIndices,
+                                    const int* resultIndices,
+                                    int matrixCount);
+
     int updateTransitionMatrices(int eigenIndex,
                                  const int* probabilityIndices,
                                  const int* firstDerivativeIndices,
@@ -293,6 +308,13 @@ public:
                                  const double* edgeLengths,
                                  int count);
 
+    int updateTransitionMatricesWithModelCategories(int* eigenIndices,
+                                 const int* probabilityIndices,
+                                 const int* firstDerivativeIndices,
+                                 const int* secondDerivativeIndices,
+                                 const double* edgeLengths,
+                                 int count);
+
     int updateTransitionMatricesWithMultipleModels(const int* eigenIndices,
                                                    const int* categoryRateIndices,
                                                    const int* probabilityIndices,
@@ -300,17 +322,48 @@ public:
                                                    const int* secondDerivativeIndices,
                                                    const double* edgeLengths,
                                                    int count);
-    
+
     int updatePartials(const int* operations,
                        int operationCount,
                        int cumulativeScalingIndex);
 
+    int updatePrePartials(const int *operations,
+                          int count,
+                          int cumulativeScaleIndex);
+
+	int calculateEdgeDerivative(const int *postBufferIndices,
+								const int *preBufferIndices,
+								const int rootBufferIndex,
+								const int *firstDerivativeIndices,
+								const int *secondDerivativeIndices,
+								const int categoryWeightsIndex,
+								const int categoryRatesIndex,
+								const int stateFrequenciesIndex,
+								const int *cumulativeScaleIndices,
+								int count,
+								double *outFirstDerivative,
+								double *outDiagonalSecondDerivative);
+
+    int calculateEdgeDerivatives(const int *postBufferIndices,
+                                 const int *preBufferIndices,
+                                 const int *derivativeMatrixIndices,
+                                 const int *categoryWeightsIndices,
+                                 const int *categoryRatesIndices,
+                                 const int *cumulativeScaleIndices,
+                                 int count,
+                                 double *outDerivatives,
+                                 double *outSumDerivatives,
+                                 double *outSumSquaredDerivatives);
+
     int updatePartialsByPartition(const int* operations,
                                   int operationCount);
-    
+
+    int updatePrePartialsByPartition(const int* operations,
+                                     int operationCount);
+
     int waitForPartials(const int* destinationPartials,
                         int destinationPartialsCount);
-    
+
     int accumulateScaleFactors(const int* scalingIndices,
                                int count,
                                int cumulativeScalingIndex);
@@ -319,7 +372,7 @@ public:
                                           int count,
                                           int cumulativeScalingIndex,
                                           int partitionIndex);
-    
+
     int removeScaleFactors(const int* scalingIndices,
                            int count,
                            int cumulativeScalingIndex);
@@ -328,17 +381,17 @@ public:
                                       int count,
                                       int cumulativeScalingIndex,
                                       int partitionIndex);
-    
+
     int resetScaleFactors(int cumulativeScalingIndex);
 
     int resetScaleFactorsByPartition(int cumulativeScalingIndex, int partitionIndex);
-    
+
     int copyScaleFactors(int destScalingIndex,
                          int srcScalingIndex);
-                         
+
     int getScaleFactors(int srcScalingIndex,
-                        double* scaleFactors);                          
-    
+                        double* scaleFactors);
+
     int calculateRootLogLikelihoods(const int* bufferIndices,
                                     const int* categoryWeightsIndices,
                                     const int* stateFrequenciesIndices,
@@ -355,7 +408,7 @@ public:
                                                int count,
                                                double* outSumLogLikelihoodByPartition,
                                                double* outSumLogLikelihood);
-    
+
     int calculateEdgeLogLikelihoods(const int* parentBufferIndices,
                                     const int* childBufferIndices,
                                     const int* probabilityIndices,
@@ -387,13 +440,22 @@ public:
                                                double* outSumSecondDerivativeByPartition,
                                                double* outSumSecondDerivative);
 
+	int calculateCrossProducts(const int *postBufferIndices,
+							   const int *preBufferIndices,
+							   const int *categoryRatesIndices,
+							   const int *categoryWeightsIndices,
+							   const double *edgeLengths,
+							   int count,
+							   double *outSumDerivatives,
+							   double *outSumSquaredDerivatives);
+
     int getLogLikelihood(double* outSumLogLikelihood);
 
     int getDerivatives(double* outSumFirstDerivative,
                        double* outSumSecondDerivative);
 
     int getSiteLogLikelihoods(double* outLogLikelihoods);
-    
+
     int getSiteDerivatives(double* outFirstDerivatives,
                            double* outSecondDerivatives);
 
@@ -405,11 +467,39 @@ private:
 
     int  reorderPatternsByPartition();
 
+    std::vector<int> transposeTransitionMatricesOnTheFly(const int *operations,
+                                                         int operationCount);
+
+    int setMatrixBufferImpl(int matrixIndex,
+                            const double* inMatrix,
+                            double paddedValue,
+                            bool transpose);
+
     int upPartials(bool byPartition,
                    const int* operations,
                    int operationCount,
                    int cumulativeScalingIndex);
 
+	int upPrePartials(bool byPartition,
+					  const int* operations,
+					  int count,
+					  int cumulativeScaleIndex);
+
+	void initDerivatives(int replicates);
+
+	int calcEdgeFirstDerivatives(const int *postBufferIndices, const int *preBufferIndices, const int *firstDerivativeIndices,
+								 const int *categoryWeightsIndices, const int *scaleIndices, int count,
+								 double *outFirstDerivatives,
+								 double *outSumFirstDerivatives,
+								 double *outSumSquaredFirstDerivatives);
+
+	int calcCrossProducts(const int *postBufferIndices,
+						  const int *preBufferIndices,
+						  const int *categoryRateIndices,
+						  const int *categoryWeightIndices,
+						  const double* edgeLengths,
+						  int totalCount,
+						  double *outCrossProducts);
 };
 
 BEAGLE_GPU_TEMPLATE
