@@ -4,19 +4,9 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @brief GPU kernel launcher
  *
@@ -44,8 +34,9 @@ class KernelLauncher {
 
 private:
     GPUInterface* gpu;
-    
+
     GPUFunction fMatrixConvolution;
+    GPUFunction fMatrixTranspose;
     GPUFunction fMatrixMulADBMulti;
     GPUFunction fMatrixMulADB;
     GPUFunction fMatrixMulADBFirstDeriv;
@@ -60,6 +51,14 @@ private:
     GPUFunction fPartialsPartialsByPatternBlockAutoScaling;
     GPUFunction fPartialsPartialsByPatternBlockCheckScaling;
     GPUFunction fPartialsPartialsByPatternBlockFixedCheckScaling;
+	GPUFunction fPartialsPartialsEdgeFirstDerivatives;
+	GPUFunction fPartialsStatesEdgeFirstDerivatives;
+	GPUFunction fPartialsPartialsCrossProducts;
+	GPUFunction fPartialsStatesCrossProducts;
+	GPUFunction fMultipleNodeSiteReduction;
+	GPUFunction fMultipleNodeSiteSquaredReduction;
+	GPUFunction fPartialsPartialsGrowing;
+	GPUFunction fPartialsStatesGrowing;
     GPUFunction fStatesPartialsByPatternBlockCoherentMulti;
     GPUFunction fStatesPartialsByPatternBlockCoherentPartition;
     GPUFunction fStatesPartialsByPatternBlockCoherent;
@@ -78,7 +77,7 @@ private:
     GPUFunction fStatesPartialsEdgeLikelihoods;
     GPUFunction fStatesPartialsEdgeLikelihoodsByPartition;
     GPUFunction fStatesPartialsEdgeLikelihoodsSecondDeriv;
-        
+
     GPUFunction fIntegrateLikelihoodsDynamicScaling;
     GPUFunction fIntegrateLikelihoodsDynamicScalingPartition;
     GPUFunction fIntegrateLikelihoodsDynamicScalingSecondDeriv;
@@ -108,11 +107,13 @@ private:
     GPUFunction fSumSites3;
 
     GPUFunction fReorderPatterns;
-    
+
     Dim3Int bgTransitionProbabilitiesBlock;
     Dim3Int bgTransitionProbabilitiesGrid;
     Dim3Int bgPeelingBlock;
     Dim3Int bgPeelingGrid;
+	Dim3Int bgDerivativeBlock;
+	Dim3Int bgDerivativeGrid;
     Dim3Int bgLikelihoodBlock;
     Dim3Int bgLikelihoodGrid;
     Dim3Int bgAccumulateBlock;
@@ -123,27 +124,29 @@ private:
     Dim3Int bgSumSitesGrid;
     Dim3Int bgReorderPatternsBlock;
     Dim3Int bgReorderPatternsGrid;
+	Dim3Int bgMultiNodeSumBlock;
+	Dim3Int bgMultiNodeSumGrid;
+
 
-    
     unsigned int kPaddedStateCount;
     unsigned int kCategoryCount;
     unsigned int kPatternCount;
     unsigned int kUnpaddedPatternCount;
     unsigned int kPatternBlockSize;
     unsigned int kMatrixBlockSize;
-    unsigned int kSlowReweighing;  
+    unsigned int kSlowReweighing;
     unsigned int kMultiplyBlockSize;
     unsigned int kSumSitesBlockSize;
     long kFlags;
     bool kCPUImplementation;
     bool kAppleCPUImplementation;
 
-    
+
 public:
     KernelLauncher(GPUInterface* inGpu);
-    
+
     ~KernelLauncher();
-    
+
     // void SetupPartitioningKernelGrid(unsigned int partitionBlockCount);
 
 // Kernel links
@@ -164,6 +167,10 @@ public:
                           GPUPtr dPtrQueue,
                           unsigned int totalMatrixCount);
 
+    void TransposeTransitionMatrices(GPUPtr dMatrices,
+    								 GPUPtr dPtrQueue,
+    								 unsigned int totalMatrixCount);
+
     void GetTransitionProbabilitiesSquareMulti(GPUPtr dMatrices,
                                                GPUPtr dPtrQueue,
                                                GPUPtr dEvec,
@@ -186,8 +193,8 @@ public:
                                                      GPUPtr dIevc,
                                                      GPUPtr dEigenValues,
                                                      GPUPtr distanceQueue,
-                                                     unsigned int totalMatrix);    
-    
+                                                     unsigned int totalMatrix);
+
     void GetTransitionProbabilitiesSquareSecondDeriv(GPUPtr dMatrices,
                                                      GPUPtr dPtrQueue,
                                           GPUPtr dEvec,
@@ -236,7 +243,7 @@ public:
                                                int doRescaling,
                                                int streamIndex,
                                                int waitIndex);
-    
+
     void StatesPartialsPruningMulti(GPUPtr states,
                                     GPUPtr partials,
                                     GPUPtr matrices,
@@ -261,7 +268,7 @@ public:
                                              int doRescaling,
                                              int streamIndex,
                                              int waitIndex);
-    
+
     void StatesStatesPruningMulti(GPUPtr states,
                                   GPUPtr partials,
                                   GPUPtr matrices,
@@ -286,7 +293,93 @@ public:
                                            int doRescaling,
                                            int streamIndex,
                                            int waitIndex);
-    
+
+	void PartialsStatesGrowing(GPUPtr partials1,
+                               GPUPtr partials2,
+                               GPUPtr partials3,
+                               GPUPtr matrices1,
+                               GPUPtr matrices2,
+                               unsigned int patternCount,
+                               unsigned int categoryCount,
+                               int sizeReal);
+
+    void PartialsPartialsGrowing(GPUPtr partials1,
+                                 GPUPtr partials2,
+                                 GPUPtr partials3,
+                                 GPUPtr matrices1,
+                                 GPUPtr matrices2,
+                                 unsigned int patternCount,
+                                 unsigned int categoryCount,
+                                 int sizeReal);
+
+	void PartialsStatesEdgeFirstDerivatives(GPUPtr out,
+											  GPUPtr states0,
+											  GPUPtr partials0,
+											  GPUPtr matrices0,
+											  GPUPtr weights,
+											  GPUPtr instructions,
+											  unsigned int instructionOffset,
+											  unsigned int nodeCount,
+											  unsigned int patternCount,
+											  unsigned int categoryCount,
+											  bool synchronize);
+
+	void PartialsPartialsEdgeFirstDerivatives(GPUPtr out,
+											  GPUPtr partials0,
+											  GPUPtr matrices0,
+											  GPUPtr weights,
+											  GPUPtr instructions,
+											  unsigned int instructionOffset,
+											  unsigned int nodeCount,
+											  unsigned int patternCount,
+											  unsigned int categoryCount,
+											  bool synchronize);
+
+	void PartialsStatesCrossProducts(GPUPtr out,
+									 GPUPtr states0,
+                                                   GPUPtr partials,
+                                                   GPUPtr lengths,
+                                                   GPUPtr instructions,
+                                                   GPUPtr categoryWeights,
+                                                   GPUPtr patternWeights,
+                                                   unsigned int instructionOffset,
+                                                   unsigned int nodeCount,
+                                                   unsigned int rateOffset,
+                                                   unsigned int patternCount,
+                                                   unsigned int categoryCount,
+                                                   bool synchronize,
+                                                   unsigned int nodeBlocks,
+                                                   unsigned int patterBlocks);
+
+	void PartialsPartialsCrossProducts(GPUPtr out,
+                                                   GPUPtr partials,
+                                                   GPUPtr lengths,
+                                                   GPUPtr instructions,
+                                                   GPUPtr categoryWeights,
+                                                   GPUPtr patternWeights,
+                                                   unsigned int instructionOffset,
+                                                   unsigned int nodeCount,
+                                                   unsigned int rateOffset,
+                                                   unsigned int patternCount,
+                                                   unsigned int categoryCount,
+                                                   bool synchronize,
+                                                   unsigned int nodeBlocks,
+                                                   unsigned int patterBlocks);
+
+	void MultipleNodeSiteReduction(GPUPtr outSiteValues,
+								   GPUPtr inSiteValues,
+								   GPUPtr weights,
+								   unsigned int outOffset,
+								   unsigned int stride,
+								   unsigned int count);
+
+	void MultipleNodeSiteSquaredReduction(GPUPtr outSiteValues,
+								          GPUPtr inSiteValues,
+								          GPUPtr weights,
+								          unsigned int outOffset,
+								          unsigned int stride,
+								          unsigned int count);
+
     void IntegrateLikelihoodsDynamicScaling(GPUPtr dResult,
                                             GPUPtr dRootPartials,
                                             GPUPtr dWeights,
@@ -304,7 +397,7 @@ public:
                                                      unsigned int patternCount,
                                                      unsigned int categoryCount,
                                                      int gridSize);
-    
+
     void IntegrateLikelihoodsAutoScaling(GPUPtr dResult,
                                             GPUPtr dRootPartials,
                                             GPUPtr dWeights,
@@ -312,7 +405,7 @@ public:
                                             GPUPtr dRootScalingFactors,
                                             unsigned int patternCount,
                                             unsigned int categoryCount);
-    
+
     void IntegrateLikelihoodsDynamicScalingSecondDeriv(GPUPtr dResult,
                                                        GPUPtr dFirstDerivResult,
                                                        GPUPtr dSecondDerivResult,
@@ -324,7 +417,7 @@ public:
                                                        GPUPtr dRootScalingFactors,
                                                        unsigned int patternCount,
                                                        unsigned int categoryCount);
-    
+
     void PartialsPartialsEdgeLikelihoods(GPUPtr dPartialsTmp,
                                          GPUPtr dParentPartials,
                                          GPUPtr dChildParials,
@@ -338,7 +431,7 @@ public:
                                                     GPUPtr dPtrOffsets,
                                                     unsigned int patternCount,
                                                     int gridSize);
-    
+
     void PartialsPartialsEdgeLikelihoodsSecondDeriv(GPUPtr dPartialsTmp,
                                                     GPUPtr dFirstDerivTmp,
                                                     GPUPtr dSecondDerivTmp,
@@ -349,8 +442,8 @@ public:
                                                     GPUPtr dSecondDerivMatrix,
                                                     unsigned int patternCount,
                                                     unsigned int categoryCount);
-    
-    
+
+
     void StatesPartialsEdgeLikelihoods(GPUPtr dPartialsTmp,
                                        GPUPtr dParentPartials,
                                        GPUPtr dChildStates,
@@ -365,7 +458,7 @@ public:
                                                   GPUPtr dPtrOffsets,
                                                   unsigned int patternCount,
                                                   int gridSize);
-    
+
     void StatesPartialsEdgeLikelihoodsSecondDeriv(GPUPtr dPartialsTmp,
                                                   GPUPtr dFirstDerivTmp,
                                                   GPUPtr dSecondDerivTmp,
@@ -376,7 +469,7 @@ public:
                                                   GPUPtr dSecondDerivMatrix,
                                                   unsigned int patternCount,
                                                   unsigned int categoryCount);
-    
+
     void AccumulateFactorsDynamicScaling(GPUPtr dScalingFactors,
                                          GPUPtr dNodePtrQueue,
                                          GPUPtr dRootScalingFactors,
@@ -396,12 +489,12 @@ public:
                                       unsigned int nodeCount,
                                       unsigned int patternCount,
                                       unsigned int scaleBufferSize);
-    
+
     void RemoveFactorsDynamicScaling(GPUPtr dScalingFactors,
                                      GPUPtr dNodePtrQueue,
                                      GPUPtr dRootScalingFactors,
                                      unsigned int nodeCount,
-                                     unsigned int patternCount);    
+                                     unsigned int patternCount);
 
     void RemoveFactorsDynamicScalingByPartition(GPUPtr dScalingFactors,
                                                 GPUPtr dNodePtrQueue,
@@ -409,7 +502,7 @@ public:
                                                 unsigned int nodeCount,
                                                 int startPattern,
                                                 int endPattern);
-    
+
     void ResetFactorsDynamicScalingByPartition(GPUPtr dScalingFactors,
                                                int startPattern,
                                                int endPattern);
@@ -440,7 +533,7 @@ public:
                               GPUPtr dFrequencies,
                               unsigned int patternCount,
                               unsigned int categoryCount);
-    
+
     void IntegrateLikelihoodsPartition(GPUPtr dResult,
                                        GPUPtr dRootPartials,
                                        GPUPtr dWeights,
@@ -460,7 +553,7 @@ public:
                                          GPUPtr dFrequencies,
                                          unsigned int patternCount,
                                          unsigned int categoryCount);
-    
+
 	void IntegrateLikelihoodsMulti(GPUPtr dResult,
 								   GPUPtr dRootPartials,
 								   GPUPtr dWeights,
@@ -468,7 +561,7 @@ public:
 								   unsigned int patternCount,
 								   unsigned int categoryCount,
 								   unsigned int takeLog);
-	
+
 	void IntegrateLikelihoodsFixedScaleMulti(GPUPtr dResult,
 											 GPUPtr dRootPartials,
 											 GPUPtr dWeights,
@@ -481,7 +574,7 @@ public:
 											 unsigned int categoryCount,
 											 unsigned int subsetCount,
 											 unsigned int subsetIndex);
-    
+
     void SumSites1(GPUPtr dArray1,
                   GPUPtr dSum1,
                   GPUPtr dPatternWeights,
@@ -493,14 +586,14 @@ public:
                             int startPattern,
                             int endPattern,
                             int blockCount);
-    
+
     void SumSites2(GPUPtr dArray1,
                   GPUPtr dSum1,
                   GPUPtr dArray2,
                   GPUPtr dSum2,
                   GPUPtr dPatternWeights,
                   unsigned int patternCount);
-    
+
     void SumSites3(GPUPtr dArray1,
                   GPUPtr dSum1,
                   GPUPtr dArray2,
@@ -509,9 +602,9 @@ public:
                   GPUPtr dSum3,
                   GPUPtr dPatternWeights,
                   unsigned int patternCount);
-	
+
     void SetupKernelBlocksAndGrids();
-    
+
 protected:
     void LoadKernels();
 
