@@ -1,6 +1,7 @@
-#ifdef _WIN32
-#include "libhmsbeagle/JNI/winjni.h"
-#endif
+
+// #ifdef _WIN32
+// #include "libhmsbeagle/JNI/winjni.h"
+// #endif
 
 #include <cstdio>
 #include <cstdlib>
@@ -46,7 +47,7 @@ JNIEXPORT jobjectArray JNICALL Java_beagle_BeagleJNIWrapper_getResourceList
 	if (rl == NULL) {
 	    return NULL;
     }
-    
+
 	jclass objClass = env->FindClass("beagle/ResourceDetails");
 	if (objClass == NULL) {
 		printf("NULL returned in FindClass: can't find class: beagle/ResourceDetails\n");
@@ -92,7 +93,7 @@ JNIEXPORT jobjectArray JNICALL Java_beagle_BeagleJNIWrapper_getResourceList
 
         env->SetObjectArrayElement(resourceArray, i, resourceObj);
 	}
-	
+
 	return resourceArray;
 }
 
@@ -126,7 +127,7 @@ JNIEXPORT jobjectArray JNICALL Java_beagle_BeagleJNIWrapper_getBenchmarkedResour
     if (brl == NULL) {
         return NULL;
     }
-    
+
     jclass objClass = env->FindClass("beagle/BenchmarkedResourceDetails");
     if (objClass == NULL) {
         printf("NULL returned in FindClass: can't find class: beagle/BenchmarkedResourceDetails\n");
@@ -229,7 +230,7 @@ JNIEXPORT jobjectArray JNICALL Java_beagle_BeagleJNIWrapper_getBenchmarkedResour
 
         env->SetObjectArrayElement(benchmarkedResourceArray, i, benchmarkedResourceObj);
     }
-    
+
     return benchmarkedResourceArray;
 }
 
@@ -242,7 +243,7 @@ JNIEXPORT jobjectArray JNICALL Java_beagle_BeagleJNIWrapper_getBenchmarkedResour
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_createInstance
     (JNIEnv *env, jobject obj, jint tipCount, jint partialsBufferCount, jint compactBufferCount,
     jint stateCount, jint patternCount, jint eigenBufferCount, jint matrixBufferCount, jint
-     categoryCount, jint scaleBufferCount, 
+     categoryCount, jint scaleBufferCount,
 	 jintArray inResourceList, jint resourceCount, jlong preferenceFlags, jlong requirementFlags,
 	 jobject outInstanceDetails)
 {
@@ -251,7 +252,7 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_createInstance
     jint *resourceList = NULL;
     if (inResourceList != NULL)
         resourceList = env->GetIntArrayElements(inResourceList, NULL);
-    
+
      jint instance = (jint)beagleCreateInstance(tipCount,
                                     partialsBufferCount,
                                     compactBufferCount,
@@ -266,29 +267,29 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_createInstance
                                     preferenceFlags,
                                     requirementFlags,
 									&instanceDetails);
-    
+
     if(inResourceList != NULL)
         env->ReleaseIntArrayElements(inResourceList, resourceList, JNI_ABORT);
-    
+
 	if (instance >= 0) {
 		jclass objClass = env->FindClass("beagle/InstanceDetails");
 		if (objClass == NULL) {
 			printf("NULL returned in FindClass: can't find class: beagle/InstanceDetails\n");
 			return BEAGLE_ERROR_GENERAL;
 		}
-		
+
 		jmethodID setResourceNumberMethodID = env->GetMethodID(objClass, "setResourceNumber", "(I)V");
 		if (setResourceNumberMethodID == NULL) {
 			printf("NULL returned in FindClass: can't find 'setResourceNumber' method in class: beagle/InstanceDetails\n");
 			return BEAGLE_ERROR_GENERAL;
 		}
-		
+
 		jmethodID setFlagsMethodID = env->GetMethodID(objClass, "setFlags", "(J)V");
 		if (setFlagsMethodID == NULL) {
 			printf("NULL returned in FindClass: can't find 'setFlags' method in class: beagle/InstanceDetails\n");
 			return BEAGLE_ERROR_GENERAL;
 		}
-		
+
 		env->CallVoidMethod(outInstanceDetails, setResourceNumberMethodID, instanceDetails.resourceNumber);
 		env->CallVoidMethod(outInstanceDetails, setFlagsMethodID, instanceDetails.flags);
 	}
@@ -326,12 +327,12 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setCPUThreadCount
  * Signature: (I[D)I
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setPatternWeights
-(JNIEnv *env, jobject obj, jint instance, jdoubleArray inPatternWeights) 
+(JNIEnv *env, jobject obj, jint instance, jdoubleArray inPatternWeights)
 {
     jdouble *patternWeights = env->GetDoubleArrayElements(inPatternWeights, NULL);
-    
+
 	jint errCode = (jint)beagleSetPatternWeights(instance, (double *)patternWeights);
-    
+
     env->ReleaseDoubleArrayElements(inPatternWeights, patternWeights, JNI_ABORT);
     return errCode;
 }
@@ -362,9 +363,9 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setTipStates
 (JNIEnv *env, jobject obj, jint instance, jint tipIndex, jintArray inTipStates)
 {
     jint *tipStates = env->GetIntArrayElements(inTipStates, NULL);
-    
+
 	jint errCode = (jint)beagleSetTipStates(instance, tipIndex, (int *)tipStates);
-    
+
     env->ReleaseIntArrayElements(inTipStates, tipStates, JNI_ABORT);
     return errCode;
 }
@@ -378,12 +379,12 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_getTipStates
 (JNIEnv *env, jobject obj, jint instance, jint tipIndex, jintArray outTipStates)
 {
 	fprintf(stderr,"beagleGetTipStates is not yet implemented\n");
-    exit(0);   
-    
-//    jint *tipStates = env->GetIntArrayElements(outTipStates, NULL);    
-// 
+    exit(0);
+
+//    jint *tipStates = env->GetIntArrayElements(outTipStates, NULL);
+//
 //	jint errCode = (jint)beagleGetTipStates(instance, tipIndex, (int *)tipStates);
-//    
+//
 //    env->ReleaseIntArrayElements(outTipStates, tipStates, 0);
 //    return errCode;
 }
@@ -397,13 +398,32 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setTipPartials
 (JNIEnv *env, jobject obj, jint instance, jint tipIndex, jdoubleArray inPartials)
 {
     jdouble *partials = env->GetDoubleArrayElements(inPartials, NULL);
-    
+
 	jint errCode = (jint)beagleSetTipPartials(instance, tipIndex, (double *)partials);
-    
+
     env->ReleaseDoubleArrayElements(inPartials, partials, JNI_ABORT);
     return errCode;
 }
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    setRootPrePartials
+ * Signature: (I[I[II)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setRootPrePartials
+(JNIEnv *env, jobject obj, jint instance, jintArray inbufferIndices, jintArray instateFrequenciesIndices, jint count)
+{
+    jint * bufferIndices = env->GetIntArrayElements(inbufferIndices, NULL);
+    jint * stateFrequenciesIndices = env->GetIntArrayElements(instateFrequenciesIndices, NULL);
+
+    jint errCode = (jint)beagleSetRootPrePartials(instance, (int *)bufferIndices, (int *)stateFrequenciesIndices, count);
+
+    env->ReleaseIntArrayElements(inbufferIndices, bufferIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(instateFrequenciesIndices, stateFrequenciesIndices, JNI_ABORT);
+    return errCode;
+}
+
+
 /*
  * Class:     beagle_BeagleJNIWrapper
  * Method:    setPartials
@@ -437,6 +457,144 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_getPartials
     return errCode;
 }
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    calculateEdgeDifferentials
+ * Signature: (I[I[I[I[II[D[D[D)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateEdgeDifferentials
+        (JNIEnv *env, jobject obj, jint instance,
+                jintArray postBufferIndices,
+                jintArray preBufferIndices,
+                jintArray derivativeMatrixIndices,
+                jintArray categoryWeightsIndices,
+                jint count,
+                jdoubleArray outDerivatives,
+                jdoubleArray outSumDerivatives,
+                jdoubleArray outSumSquaredDerivatives) {
+
+    jdouble *jOutDerivatives = outDerivatives == NULL ?
+            NULL : env->GetDoubleArrayElements(outDerivatives, NULL);
+
+    jdouble* jOutSumDerivatives = outSumDerivatives == NULL ?
+            NULL : env->GetDoubleArrayElements(outSumDerivatives, NULL);
+
+    jdouble* jOutSumSquaredDerivatives = outSumSquaredDerivatives == NULL ?
+            NULL : env->GetDoubleArrayElements(outSumSquaredDerivatives, NULL);
+
+    jint *jPostBufferIndices = env->GetIntArrayElements(postBufferIndices, NULL);
+    jint *jPreBufferIndices  = env->GetIntArrayElements(preBufferIndices, NULL);
+
+    jint *jDerivativeMatrixIndices = env->GetIntArrayElements(derivativeMatrixIndices, NULL);
+    jint *jCategoryWeightsIndices = env->GetIntArrayElements(categoryWeightsIndices, NULL);
+
+    jint errCode = beagleCalculateEdgeDerivatives(instance, (int*)jPostBufferIndices, (int*)jPreBufferIndices,
+                                                 (int*)jDerivativeMatrixIndices, (int*)jCategoryWeightsIndices, count,
+                                                 jOutDerivatives, jOutSumDerivatives, jOutSumSquaredDerivatives);
+
+    env->ReleaseIntArrayElements(postBufferIndices, jPostBufferIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(preBufferIndices, jPreBufferIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(derivativeMatrixIndices, jDerivativeMatrixIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(categoryWeightsIndices, jCategoryWeightsIndices, JNI_ABORT);
+
+    if (outDerivatives != NULL) {
+        env->ReleaseDoubleArrayElements(outDerivatives, jOutDerivatives, 0);
+    }
+
+    if (outSumDerivatives != NULL) {
+        env->ReleaseDoubleArrayElements(outSumDerivatives, jOutSumDerivatives, 0);
+    }
+
+    if (outSumSquaredDerivatives != NULL) {
+        env->ReleaseDoubleArrayElements(outSumSquaredDerivatives, jOutSumSquaredDerivatives, 0);
+    }
+
+    return errCode;
+}
+
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateCrossProductDifferentials
+  (JNIEnv *env, jobject obj, jint instance,
+		jintArray postBufferIndices,
+		jintArray preBufferIndices,
+		jintArray categoryRatesIndices,
+		jintArray categoryWeightsIndices,
+		jdoubleArray edgeLengths,
+		jint count,
+		jdoubleArray outSumDerivatives,
+		jdoubleArray outSumSquaredDerivatives) {
+
+    jdouble* jOutSumDerivatives = outSumDerivatives == NULL ?
+            NULL : env->GetDoubleArrayElements(outSumDerivatives, NULL);
+
+    jdouble* jOutSumSquaredDerivatives = outSumSquaredDerivatives == NULL ?
+            NULL : env->GetDoubleArrayElements(outSumSquaredDerivatives, NULL);
+
+    jint *jPostBufferIndices = env->GetIntArrayElements(postBufferIndices, NULL);
+    jint *jPreBufferIndices  = env->GetIntArrayElements(preBufferIndices, NULL);
+
+    jint *jCategoryRatesIndices = env->GetIntArrayElements(categoryRatesIndices, NULL);
+    jint *jCategoryWeightsIndices = env->GetIntArrayElements(categoryWeightsIndices, NULL);
+
+    jdouble* jEdgeLengths = env->GetDoubleArrayElements(edgeLengths, NULL);
+
+    jint errCode = beagleCalculateCrossProductDerivative(instance, (int*)jPostBufferIndices, (int*)jPreBufferIndices,
+                                                         (int*)jCategoryRatesIndices, (int*)jCategoryWeightsIndices,
+                                                         (double*)jEdgeLengths, count,
+                                                         jOutSumDerivatives, jOutSumSquaredDerivatives);
+
+    env->ReleaseIntArrayElements(postBufferIndices, jPostBufferIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(preBufferIndices, jPreBufferIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(categoryRatesIndices, jCategoryRatesIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(categoryWeightsIndices, jCategoryWeightsIndices, JNI_ABORT);
+    env->ReleaseDoubleArrayElements(edgeLengths, jEdgeLengths, JNI_ABORT);
+
+    if (outSumDerivatives != NULL) {
+        env->ReleaseDoubleArrayElements(outSumDerivatives, jOutSumDerivatives, 0);
+    }
+
+    if (outSumSquaredDerivatives != NULL) {
+        env->ReleaseDoubleArrayElements(outSumSquaredDerivatives, jOutSumSquaredDerivatives, 0);
+    }
+
+    return errCode;
+
+
+}
+
+
+
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    calculateEdgeDerivative
+ * Signature: (I[I[II[I[IIII[II[D[D)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateEdgeDerivative
+(JNIEnv *env, jobject obj, jint instance, jintArray postBufferIndices, jintArray preBufferIndices,
+ jint rootBufferIndex, jintArray firstDerivativeIndices, jintArray secondDerivativeIndices,
+ jint categoryWeightsIndex, jint categoryRatesIndex, jint stateFrequenciesIndex,
+ jintArray cumulativeScaleIndices, jint count,
+ jdoubleArray outFirstDerivative, jdoubleArray outDiagonalSecondDerivative)
+{
+    jdouble *outGradient = env->GetDoubleArrayElements(outFirstDerivative, NULL);
+    jdouble *outDiagonalHessian = outDiagonalSecondDerivative != NULL ? env->GetDoubleArrayElements(outDiagonalSecondDerivative, NULL) : NULL;
+    jint *postOrderBufferIndices = env->GetIntArrayElements(postBufferIndices, NULL);
+    jint *preOrderBufferIndices  = env->GetIntArrayElements(preBufferIndices, NULL);
+    jint *gradientMatrixIndices = env->GetIntArrayElements(firstDerivativeIndices, NULL);
+    jint *hessianMatrixIndices  = secondDerivativeIndices != NULL ? env->GetIntArrayElements(secondDerivativeIndices, NULL) : NULL;
+    jint *cumulativeScalingIndices = env->GetIntArrayElements(cumulativeScaleIndices, NULL);
+
+
+    jint errCode = beagleCalculateEdgeDerivative(instance, (int*)postOrderBufferIndices, (int*)preOrderBufferIndices,
+                                                 rootBufferIndex, (int*)gradientMatrixIndices, (int*)hessianMatrixIndices,
+                                                 categoryWeightsIndex, categoryRatesIndex, stateFrequenciesIndex,
+                                                 (int*)cumulativeScalingIndices, count, outGradient, outDiagonalHessian);
+    env->ReleaseDoubleArrayElements(outFirstDerivative, outGradient, 0);
+    if (outDiagonalSecondDerivative != NULL) {
+        env->ReleaseDoubleArrayElements(outDiagonalSecondDerivative, outDiagonalHessian, 0);
+    }
+    return errCode;
+}
+
 /*
  * Class:     beagle_BeagleJNIWrapper
  * Method:    getLogScaleFactors
@@ -485,11 +643,11 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setStateFrequencies
 (JNIEnv *env, jobject obj, jint instance, jint stateFrequenciesIndex, jdoubleArray inStateFrequencies)
 {
     jdouble *stateFrequencies = env->GetDoubleArrayElements(inStateFrequencies, NULL);
-	
+
 	jint errCode = (jint)beagleSetStateFrequencies(instance, stateFrequenciesIndex, (double *)stateFrequencies);
-	
+
     env->ReleaseDoubleArrayElements(inStateFrequencies, stateFrequencies, JNI_ABORT);
-	
+
     return errCode;
 }
 
@@ -502,11 +660,11 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setCategoryWeights
 (JNIEnv *env, jobject obj, jint instance, jint categoryWeightsIndex, jdoubleArray inCategoryWeights)
 {
     jdouble *categoryWeights = env->GetDoubleArrayElements(inCategoryWeights, NULL);
-	
+
 	jint errCode = (jint)beagleSetCategoryWeights(instance, categoryWeightsIndex, (double *)categoryWeights);
-	
+
     env->ReleaseDoubleArrayElements(inCategoryWeights, categoryWeights, JNI_ABORT);
-	
+
     return errCode;
 }
 
@@ -563,6 +721,23 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setTransitionMatrix
     return errCode;
 }
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    setDifferentialMatrix
+ * Signature: (II[D)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_setDifferentialMatrix
+  (JNIEnv *env, jobject obj, jint instance, jint matrixIndex, jdoubleArray inMatrix)
+{
+    jdouble *matrix = env->GetDoubleArrayElements(inMatrix, NULL);
+
+	jint errCode = (jint)beagleSetDifferentialMatrix(instance, matrixIndex, (double *)matrix);
+
+    env->ReleaseDoubleArrayElements(inMatrix, matrix, JNI_ABORT);
+
+    return errCode;
+}
+
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_getTransitionMatrix
   (JNIEnv *env, jobject obj, jint instance, jint matrixIndex, jdoubleArray outMatrix)
 {
@@ -603,6 +778,50 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_convolveTransitionMatrices
     return errCode;
 }
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    addTransitionMatrices
+ * Signature: (I[D)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_addTransitionMatrices
+        (JNIEnv *env, jobject obj, jint instance, jintArray inFirstIndices, jintArray inSecondIndices, jintArray inResultIndices, jint matrixCount)
+{
+    jint errCode;
+
+    jint *firstIndices = env->GetIntArrayElements(inFirstIndices, NULL);
+    jint *secondIndices = env->GetIntArrayElements(inSecondIndices, NULL);
+    jint *resultIndices = env->GetIntArrayElements(inResultIndices, NULL);
+
+    errCode = (jint)beagleAddTransitionMatrices(instance, (int *)firstIndices, (int *)secondIndices, (int *)resultIndices, matrixCount);
+
+    env->ReleaseIntArrayElements(inFirstIndices, firstIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(inSecondIndices, secondIndices, JNI_ABORT);
+    env->ReleaseIntArrayElements(inResultIndices, resultIndices, JNI_ABORT);
+
+    return errCode;
+}
+
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    transposeTransitionMatrices
+ * Signature: (I[I[II)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_transposeTransitionMatrices
+        (JNIEnv *env, jobject, jint instance, jintArray jInIndices, jintArray jOutIndices, jint matrixCount) {
+
+    jint errCode;
+
+    jint *inIndices = env->GetIntArrayElements(jInIndices, NULL);
+    jint *outIndices = env->GetIntArrayElements(jOutIndices, NULL);
+
+    errCode = (jint)beagleTransposeTransitionMatrices(instance, (int *)inIndices, (int *)outIndices, matrixCount);
+
+    env->ReleaseIntArrayElements(jInIndices, inIndices, JNI_ABORT);
+
+    return errCode;
+}
+
+
 
 /*
  * Class:     beagle_BeagleJNIWrapper
@@ -652,11 +871,44 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_updateTransitionMatricesWith
     return errCode;
 }
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    updatePrePartials
+ * Signature: (I[III)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_updatePrePartials
+  (JNIEnv *env, jobject obj, jint instance, jintArray inOperations, jint operationCount, jint cumulativeScaleIndex)
+{
+    jint *operations = env->GetIntArrayElements(inOperations, NULL);
+
+    jint errCode = (jint)beagleUpdatePrePartials(instance, (BeagleOperation*)operations, operationCount, cumulativeScaleIndex);
+
+    env->ReleaseIntArrayElements(inOperations, operations, JNI_ABORT);
+
+    return errCode;
+}
+
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    updatePrePartialsByPartition
+ * Signature: (I[II)I
+ */
+JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_updatePrePartialsByPartition
+  (JNIEnv *env, jobject obj, jint instance, jintArray inOperations, jint operationCount)
+{
+    jint *operations = env->GetIntArrayElements(inOperations, NULL);
+
+    jint errCode = (jint)beagleUpdatePrePartialsByPartition(instance, (BeagleOperationByPartition*)operations, operationCount);
+
+    env->ReleaseIntArrayElements(inOperations, operations, JNI_ABORT);
+
+    return errCode;
+}
 
 /*
  * Class:     beagle_BeagleJNIWrapper
  * Method:    updatePartials
- * Signature: ([II[III)I
+ * Signature: (I[III)I
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_updatePartials
   (JNIEnv *env, jobject obj, jint instance, jintArray inOperations, jint operationCount, jint cumulativeScalingIndex)
@@ -711,7 +963,7 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_waitForPartials
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_accumulateScaleFactors
   (JNIEnv *env, jobject obj, jint instance, jintArray inScaleIndices, jint count, jint cumulativeScalingIndex) {
-	
+
 	jint *scaleIndices = env->GetIntArrayElements(inScaleIndices, NULL);
 	jint errCode = (jint)beagleAccumulateScaleFactors(instance, (int*)scaleIndices, count, cumulativeScalingIndex);
 	env->ReleaseIntArrayElements(inScaleIndices, scaleIndices, JNI_ABORT);
@@ -741,11 +993,11 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_accumulateScaleFactorsByPart
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_removeScaleFactors
 (JNIEnv *env, jobject obj, jint instance, jintArray inScaleIndices, jint count, jint cumulativeScalingIndex) {
-	
+
 	jint *scaleIndices = env->GetIntArrayElements(inScaleIndices, NULL);
 	jint errCode = (jint)beagleRemoveScaleFactors(instance, (int*)scaleIndices, count, cumulativeScalingIndex);
 	env->ReleaseIntArrayElements(inScaleIndices, scaleIndices, JNI_ABORT);
-    
+
 	return errCode;
 }
 
@@ -760,7 +1012,7 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_removeScaleFactorsByPartitio
     jint *scaleIndices = env->GetIntArrayElements(inScaleIndices, NULL);
     jint errCode = (jint)beagleRemoveScaleFactorsByPartition(instance, (int*)scaleIndices, count, cumulativeScalingIndex, partitionIndex);
     env->ReleaseIntArrayElements(inScaleIndices, scaleIndices, JNI_ABORT);
-    
+
     return errCode;
 }
 
@@ -771,7 +1023,7 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_removeScaleFactorsByPartitio
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_resetScaleFactors
 (JNIEnv *env, jobject obj, jint instance, jint cumulativeScalingIndex) {
-	
+
 	jint errCode = (jint)beagleResetScaleFactors(instance, cumulativeScalingIndex);
 	return errCode;
 }
@@ -807,7 +1059,7 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_copyScaleFactors
  * Signature: (I[I[D[DI[D)I
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateRootLogLikelihoods
-  (JNIEnv *env, jobject obj, jint instance, jintArray inBufferIndices, jintArray inCategoryWeightsIndices, 
+  (JNIEnv *env, jobject obj, jint instance, jintArray inBufferIndices, jintArray inCategoryWeightsIndices,
    jintArray inStateFrequenciesIndices, jintArray inScalingIndices, jint count, jdoubleArray outSumLogLikelihoods)
 {
     jint *bufferIndices = env->GetIntArrayElements(inBufferIndices, NULL);
@@ -815,12 +1067,12 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateRootLogLikelihoods
     jint *frequenciesIndices = env->GetIntArrayElements(inStateFrequenciesIndices, NULL);
     jint *scalingIndices = env->GetIntArrayElements(inScalingIndices, NULL);
     //    jint *scalingCount = env->GetIntArrayElements(inScalingCount, NULL);
-	
+
     jdouble *sumLogLikelihoods = env->GetDoubleArrayElements(outSumLogLikelihoods, NULL);
 
-	jint errCode = (jint)beagleCalculateRootLogLikelihoods(instance, (int *)bufferIndices, 
+	jint errCode = (jint)beagleCalculateRootLogLikelihoods(instance, (int *)bufferIndices,
 														   (int *)weightsIndices,
-														   (int *)frequenciesIndices, 
+														   (int *)frequenciesIndices,
 														   (int *)scalingIndices,
 														   count, (double *)sumLogLikelihoods);
 
@@ -851,14 +1103,14 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateRootLogLikelihoodsB
     jint *scalingIndices = env->GetIntArrayElements(inScalingIndices, NULL);
     jint *partitionIndices = env->GetIntArrayElements(inPartitionIndices, NULL);
 
-    jdouble *sumLogLikelihoodsByPartition = env->GetDoubleArrayElements(outSumLogLikelihoodByPartition, NULL);    
+    jdouble *sumLogLikelihoodsByPartition = env->GetDoubleArrayElements(outSumLogLikelihoodByPartition, NULL);
     jdouble *sumLogLikelihoods = env->GetDoubleArrayElements(outSumLogLikelihoods, NULL);
 
     jint errCode = (jint)beagleCalculateRootLogLikelihoodsByPartition(
                                                            instance,
-                                                           (int *)bufferIndices, 
+                                                           (int *)bufferIndices,
                                                            (int *)weightsIndices,
-                                                           (int *)frequenciesIndices, 
+                                                           (int *)frequenciesIndices,
                                                            (int *)scalingIndices,
                                                            (int *)partitionIndices,
                                                            partitionCount,
@@ -906,12 +1158,12 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateEdgeLogLikelihoods
 
 	jint errCode = (jint)beagleCalculateEdgeLogLikelihoods(instance, (int *)parentBufferIndices, (int *)childBufferIndices,
 	                                                    (int *)probabilityIndices, (int *)firstDerivativeIndices,
-	                                                    (int *)secondDerivativeIndices, 
+	                                                    (int *)secondDerivativeIndices,
 														(int *)weightsIndices,
-														(int *)frequenciesIndices, 
+														(int *)frequenciesIndices,
 														(int *)scalingIndices,// (int *)scalingCount,
-	                                                    count, 
-														(double *)sumLogLikelihoods, 
+	                                                    count,
+														(double *)sumLogLikelihoods,
 														(double *)sumFirstDerivatives,
 	                                                    (double *)sumSecondDerivatives);
 
@@ -942,22 +1194,22 @@ JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_calculateEdgeLogLikelihoods
  */
 JNIEXPORT jint JNICALL Java_beagle_BeagleJNIWrapper_getSiteLogLikelihoods
 (JNIEnv *env, jobject obj, jint instance, jdoubleArray outSiteLogLikelihoods) {
-	
+
 	jdouble *siteLogLikelihoods = env->GetDoubleArrayElements(outSiteLogLikelihoods, NULL);
-	
+
 	jint errCode = (jint)beagleGetSiteLogLikelihoods(instance, (double *)siteLogLikelihoods);
-	
+
     // not using JNI_ABORT flag here because we want the values to be copied back...
     env->ReleaseDoubleArrayElements(outSiteLogLikelihoods, siteLogLikelihoods, 0);
     return errCode;
 }
 
 //void __attribute__ ((constructor)) beagle_jni_library_initialize(void) {
-//	
+//
 //}
 
 
 //void __attribute__ ((destructor)) beagle_jni_library_finialize(void) {
-//	
+//
 //}
 
