@@ -6,19 +6,9 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @author Andrew Rambaut
  * @author Marc Suchard
@@ -49,11 +39,12 @@
 #include "libhmsbeagle/benchmark/BeagleBenchmark.h"
 
 #include "libhmsbeagle/plugin/Plugin.h"
+#include "beagle.h"
 
-#define BEAGLE_VERSION  PACKAGE_VERSION  //" (PRE-RELEASE)"
+#define BEAGLE_VERSION  PACKAGE_VERSION  " (PRE-RELEASE)"
 #define BEAGLE_CITATION "Using BEAGLE library v" BEAGLE_VERSION " for accelerated, parallel likelihood evaluation\n\
 2009-, BEAGLE Working Group - https://beagle-dev.github.io/\n\
-Citation: Ayres et al (2012) Systematic Biology 61: 170-173 | doi:10.1093/sysbio/syr100\n"
+Citation: Ayres et al (2019) Systematic Biology 68: 1052-1061 | doi:10.1093/sysbio/syz020\n"
 
 typedef std::pair<int, int> IntPair;
 typedef std::list<IntPair> PairedList;
@@ -61,6 +52,9 @@ typedef std::pair<int, std::pair<int, beagle::BeagleImplFactory*> > RsrcImpl;
 typedef std::list<RsrcImpl> RsrcImplList;
 typedef std::list<BeagleBenchmarkedResource> RsrcBenchPairList;
 
+// #define BEAGLE_DEBUG_LOAD
+#define BEAGLE_PREORDER
+
 //#define BEAGLE_DEBUG_TIME
 #ifdef BEAGLE_DEBUG_TIME
 #include <sys/time.h>
@@ -140,9 +134,16 @@ void beagleLoadPlugins(void) {
     beagle::plugin::PluginManager& pm = beagle::plugin::PluginManager::instance();
 
     try{
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Loading hmsbeagle-cpu-sse" << std::endl;
+#endif
         beagle::plugin::Plugin* sseplug = pm.findPlugin("hmsbeagle-cpu-sse");
         plugins->push_back(sseplug);
-    }catch(beagle::plugin::SharedLibraryException sle){}
+    }catch(beagle::plugin::SharedLibraryException sle){
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Unable to load hmsbeagle-cpu-sse: " << sle.getError() << std::endl;
+#endif
+    }
 
     try{
         beagle::plugin::Plugin* cpuplug = pm.findPlugin("hmsbeagle-cpu");
@@ -154,24 +155,38 @@ void beagleLoadPlugins(void) {
     }
 
     try{
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Loading hmsbeagle-cuda" << std::endl;
+#endif
         beagle::plugin::Plugin* cudaplug = pm.findPlugin("hmsbeagle-cuda");
         plugins->push_back(cudaplug);
-    }catch(beagle::plugin::SharedLibraryException sle){}
+    }catch(beagle::plugin::SharedLibraryException sle) {
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Unable to load hmsbeagle-cuda: " << sle.getError() << std::endl;
+#endif
+    }
 
     try{
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Loading hmsbeagle-opencl" << std::endl;
+#endif
         beagle::plugin::Plugin* openclplug = pm.findPlugin("hmsbeagle-opencl");
         plugins->push_back(openclplug);
-    }catch(beagle::plugin::SharedLibraryException sle){}
+    }catch(beagle::plugin::SharedLibraryException sle) {
+#ifdef BEAGLE_DEBUG_LOAD
+        std::cerr << "Unable to load hmsbeagle-opencl: " << sle.getError() << std::endl;
+#endif
+    }
 
     try{
         beagle::plugin::Plugin* openclalteraplug = pm.findPlugin("hmsbeagle-opencl-altera");
         plugins->push_back(openclalteraplug);
     }catch(beagle::plugin::SharedLibraryException sle){}
-    
+
     try{
         beagle::plugin::Plugin* avxplug = pm.findPlugin("hmsbeagle-cpu-avx");
         plugins->push_back(avxplug);
-    }catch(beagle::plugin::SharedLibraryException sle){}    
+    }catch(beagle::plugin::SharedLibraryException sle){}
 
     try{
         beagle::plugin::Plugin* openmpplug = pm.findPlugin("hmsbeagle-cpu-openmp");
@@ -187,7 +202,7 @@ std::list<beagle::BeagleImplFactory*>* beagleGetFactoryList(void) {
         for(; plugin_iter != plugins->end(); plugin_iter++ ){
             std::list<beagle::BeagleImplFactory*> factories = (*plugin_iter)->getBeagleFactories();
             implFactory->insert(implFactory->end(), factories.begin(), factories.end());
-        }               
+        }
     }
     return implFactory;
 }
@@ -205,7 +220,7 @@ void beagle_library_finalize(void) {
     for(; plugin_iter != plugins.end(); plugin_iter++ ){
         delete *plugin_iter;
     }
-    plugins.clear();    
+    plugins.clear();
 */
 
     if(plugins!=NULL && loaded){
@@ -291,7 +306,7 @@ BeagleResourceList* beagleGetResourceList() {
         for(; plugin_iter != plugins->end(); plugin_iter++ ){
             rsrcList->length += (*plugin_iter)->getBeagleResources().size();
         }
-        
+
         // allocate space for a complete list of resources
         rsrcList->list = (BeagleResource*) malloc(sizeof(BeagleResource) * rsrcList->length);
 
@@ -303,7 +318,7 @@ BeagleResourceList* beagleGetResourceList() {
             int prev_rI = rI;
             for(; r_iter != rList.end(); r_iter++){
                 bool rsrcExists = false;
-                for(int i=0; i<prev_rI; i++){         
+                for(int i=0; i<prev_rI; i++){
                     if (strcmp(rsrcList->list[i].name, r_iter->name) == 0) {
                         if (!rsrcExists) {
                             rsrcExists = true;
@@ -312,7 +327,7 @@ BeagleResourceList* beagleGetResourceList() {
                         rsrcList->list[i].supportFlags |= r_iter->supportFlags;
                     }
                 }
-                
+
                 if (!rsrcExists) {
                     ResourceMap.insert(std::pair<int, int>(rI, (rI - prev_rI)));
                     rsrcList->list[rI++] = *r_iter;
@@ -370,7 +385,7 @@ int filterResources(int* resourceList,
                 break;
         }
     }
-    
+
     if (possibleResources->size() == 0) {
         return BEAGLE_ERROR_NO_RESOURCE;
     }
@@ -382,21 +397,21 @@ int rankResourceImplementationPairs(long preferenceFlags,
                                     long requirementFlags,
                                     PairedList* possibleResources,
                                     RsrcImplList* possibleResourceImplementations) {
-    
+
     possibleResources->sort(compareOnFirst); // Attempt in rank order, lowest score wins
-    
+
     // Score each resource-implementation pair given preferences
 
     for(PairedList::iterator it = possibleResources->begin();
         it != possibleResources->end(); ++it) {
         int resource = (*it).second;
         long resourceRequiredFlags = rsrcList->list[resource].requiredFlags;
-        long resourceSupportedFlags = rsrcList->list[resource].supportFlags;            
+        long resourceSupportedFlags = rsrcList->list[resource].supportFlags;
         int resourceScore = (*it).first;
 #ifdef BEAGLE_DEBUG_FLOW
         fprintf(stderr,"Possible resource: %s (%d)\n",rsrcList->list[resource].name,resourceScore);
 #endif
-        
+
         for (std::list<beagle::BeagleImplFactory*>::iterator factory =
              implFactory->begin(); factory != implFactory->end(); factory++) {
             long factoryFlags = (*factory)->getFlags();
@@ -413,27 +428,27 @@ int rankResourceImplementationPairs(long preferenceFlags,
                 fprintf(stderr,"\tPossible implementation: %s (%d)\n",
                         (*factory)->getName(),totalScore);
 #endif
-                
+
                 possibleResourceImplementations->push_back(std::make_pair(totalScore, std::make_pair(resource, (*factory))));
-                
+
             }
         }
     }
-        
+
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr,"\nOriginal list of possible implementations:\n");
-    for (RsrcImplList::iterator it = possibleResourceImplementations->begin(); 
+    for (RsrcImplList::iterator it = possibleResourceImplementations->begin();
             it != possibleResourceImplementations->end(); ++it) {
         beagle::BeagleImplFactory* factory = (*it).second.second;
         fprintf(stderr,"\t %s (%d)\n", factory->getName(), (*it).first);
     }
-#endif        
-    
+#endif
+
     possibleResourceImplementations->sort(compareRsrcImpl);
-    
+
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr,"\nSorted list of possible implementations:\n");
-    for (RsrcImplList::iterator it = possibleResourceImplementations->begin(); 
+    for (RsrcImplList::iterator it = possibleResourceImplementations->begin();
             it != possibleResourceImplementations->end(); ++it) {
         beagle::BeagleImplFactory* factory = (*it).second.second;
         fprintf(stderr,"\t %s (%d)  (%d)\n", factory->getName(), (*it).first, (*it).second.first);
@@ -511,7 +526,7 @@ BeagleBenchmarkedResourceList* beagleGetBenchmarkedResourceList(int tipCount,
     double benchmarkResultCPU;
 
     bool instOnly = false;
-    
+
     errorCode = beagle::benchmark::benchmarkResource(0,
                                   stateCount,
                                   tipCount,
@@ -635,12 +650,12 @@ int beagleCreateInstance(int tipCount,
 
         if (rsrcList == NULL)
             beagleGetResourceList();
-        
+
         if (implFactory == NULL)
             beagleGetFactoryList();
-        
+
         loaded = 1;
-        
+
         int errorCode = BEAGLE_SUCCESS;
 
         PairedList* possibleResources = new PairedList;
@@ -676,7 +691,7 @@ int beagleCreateInstance(int tipCount,
         for(RsrcImplList::iterator it = possibleResourceImplementations->begin(); it != possibleResourceImplementations->end(); ++it) {
             int resource = (*it).second.first;
             beagle::BeagleImplFactory* factory = (*it).second.second;
-            
+
             bestBeagle = factory->createImpl(tipCount, partialsBufferCount,
                                                                 compactBufferCount, stateCount,
                                                                 patternCount, eigenBufferCount,
@@ -687,28 +702,29 @@ int beagleCreateInstance(int tipCount,
                                                                 preferenceFlags,
                                                                 requirementFlags,
                                                                 &errorCode);
-            
+
             if (bestBeagle != NULL)
-                break; 
+                break;
         }
-        
+
         delete possibleResourceImplementations;
-        
+
         if (bestBeagle != NULL) {
+
             int instance = instances->size();
             instances->push_back(bestBeagle);
-            
+
             int returnValue = bestBeagle->getInstanceDetails(returnInfo);
             if (returnValue == BEAGLE_SUCCESS) {
                 returnInfo->resourceName = rsrcList->list[returnInfo->resourceNumber].name;
                 // TODO: move implDescription to inside the implementation
                 returnInfo->implDescription = (char*) "none";
-                
+
                 returnValue = instance;
             }
             return returnValue;
-        }   
-        
+        }
+
         // No implementations found or appropriate, return last error code
         return errorCode;
     }
@@ -826,6 +842,32 @@ int beagleSetPartials(int instance,
     }
 }
 
+int beagleSetRootPrePartials(const int instance,
+                             const int *bufferIndices,
+                             const int *stateFrequenciesIndices,
+                             int count){
+    DEBUG_START_TIME();
+    try {
+        beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+        if (beagleInstance == NULL)
+            return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+        int returnValue = beagleInstance->setRootPrePartials(bufferIndices,
+                                                             stateFrequenciesIndices,
+                                                             count);
+        DEBUG_END_TIME();
+        return returnValue;
+    }
+    catch (std::bad_alloc &) {
+        return BEAGLE_ERROR_OUT_OF_MEMORY;
+    }
+    catch (std::out_of_range &) {
+        return BEAGLE_ERROR_OUT_OF_RANGE;
+    }
+    catch (...) {
+        return BEAGLE_ERROR_UNIDENTIFIED_EXCEPTION;
+    }
+}
+
 int beagleGetPartials(int instance, int bufferIndex, int scaleIndex, double* outPartials) {
     DEBUG_START_TIME();
     try {
@@ -978,6 +1020,19 @@ int beagleSetTransitionMatrix(int instance,
 //    }
 }
 
+int beagleSetDifferentialMatrix(int instance,
+                                int matrixIndex,
+                                const double* inMatrix) {
+    DEBUG_START_TIME();
+
+    beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+    if (beagleInstance == NULL)
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    int returnValue = beagleInstance->setDifferentialMatrix(matrixIndex, inMatrix);
+    DEBUG_END_TIME();
+    return returnValue;
+}
+
 int beagleSetTransitionMatrices(int instance,
                               const int* matrixIndices,
                               const double* inMatrices,
@@ -1015,10 +1070,6 @@ int beagleGetTransitionMatrix(int instance,
     return returnValue;
 }
 
-///////////////////////////
-//---TODO: Epoch model---//
-///////////////////////////
-
 int beagleConvolveTransitionMatrices(int instance,
                                      const int* firstIndices,
                                      const int* secondIndices,
@@ -1035,8 +1086,43 @@ int beagleConvolveTransitionMatrices(int instance,
         DEBUG_END_TIME();
         return returnValue;
     }
+}
 
-}//END: beagleConvolveTransitionMatrices
+int beagleAddTransitionMatrices(int instance,
+                                const int* firstIndices,
+                                const int* secondIndices,
+                                const int* resultIndices,
+                                int matrixCount) {
+    DEBUG_START_TIME();
+    beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+
+    if (beagleInstance == NULL) {
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    } else {
+        int returnValue = beagleInstance->addTransitionMatrices(
+                firstIndices, secondIndices,
+                resultIndices, matrixCount);
+        DEBUG_END_TIME();
+        return returnValue;
+    }
+}
+
+int beagleTransposeTransitionMatrices(int instance,
+                                      const int* inputIndices,
+                                      const int* resultIndices,
+                                      int matrixCount) {
+    DEBUG_START_TIME();
+    beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+
+    if (beagleInstance == NULL) {
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    } else {
+        int returnValue = beagleInstance->transposeTransitionMatrices(
+                inputIndices, resultIndices, matrixCount);
+        DEBUG_END_TIME();
+        return returnValue;
+    }
+}
 
 int beagleUpdateTransitionMatrices(int instance,
                              int eigenIndex,
@@ -1067,6 +1153,36 @@ int beagleUpdateTransitionMatrices(int instance,
 //    }
 }
 
+int beagleUpdateTransitionMatricesWithModelCategories(int instance,
+                             int* eigenIndices,
+                             const int* probabilityIndices,
+                             const int* firstDerivativeIndices,
+                             const int* secondDerivativeIndices,
+                             const double* edgeLengths,
+                             int count) {
+    DEBUG_START_TIME();
+//    try {
+        beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+        if (beagleInstance == NULL)
+            return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+        int returnValue = beagleInstance->updateTransitionMatricesWithModelCategories(eigenIndices, probabilityIndices,
+                                                        firstDerivativeIndices,
+                                                        secondDerivativeIndices, edgeLengths, count);
+        DEBUG_END_TIME();
+        return returnValue;
+//    }
+//    catch (std::bad_alloc &) {
+//        return BEAGLE_ERROR_OUT_OF_MEMORY;
+//    }
+//    catch (std::out_of_range &) {
+//        return BEAGLE_ERROR_OUT_OF_RANGE;
+//    }
+//    catch (...) {
+//        return BEAGLE_ERROR_UNIDENTIFIED_EXCEPTION;
+//    }
+}
+
+
 int beagleUpdateTransitionMatricesWithMultipleModels(int instance,
                                                      const int* eigenIndices,
                                                      const int* categoryRateIndices,
@@ -1088,9 +1204,9 @@ int beagleUpdateTransitionMatricesWithMultipleModels(int instance,
 
 
 int beagleUpdatePartials(const int instance,
-                   const BeagleOperation* operations,
-                   int operationCount,
-                   int cumulativeScalingIndex) {
+                         const BeagleOperation* operations,
+                         int operationCount,
+                         int cumulativeScalingIndex) {
     DEBUG_START_TIME();
 //    try {
         beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
@@ -1111,6 +1227,20 @@ int beagleUpdatePartials(const int instance,
 //    }
 }
 
+int beagleUpdatePrePartials(const int instance,
+                            const BeagleOperation* operations,
+                            int operationCount,
+                            int cumulativeScalingIndex){
+    DEBUG_START_TIME();
+    beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+    if (beagleInstance == NULL)
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    int returnValue = beagleInstance->updatePrePartials((const int *) operations, operationCount,
+                                                        cumulativeScalingIndex);
+    DEBUG_END_TIME();
+    return returnValue;
+}
+
 int beagleUpdatePartialsByPartition(const int instance,
                                     const BeagleOperationByPartition* operations,
                                     int operationCount) {
@@ -1123,6 +1253,18 @@ int beagleUpdatePartialsByPartition(const int instance,
     return returnValue;
 }
 
+int beagleUpdatePrePartialsByPartition(const int instance,
+                                       const BeagleOperationByPartition* operations,
+                                       int operationCount) {
+    DEBUG_START_TIME();
+    beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
+    if (beagleInstance == NULL)
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    int returnValue = beagleInstance->updatePrePartialsByPartition((const int*)operations, operationCount);
+    DEBUG_END_TIME();
+    return returnValue;
+}
+
 int beagleWaitForPartials(const int instance,
                     const int* destinationPartials,
                     int destinationPartialsCount) {
@@ -1148,9 +1290,9 @@ int beagleWaitForPartials(const int instance,
 }
 
 int beagleAccumulateScaleFactors(int instance,
-                           const int* scalingIndices,
-                           int count,
-                           int cumulativeScalingIndex) {
+						   const int* scalingIndices,
+						   int count,
+						   int cumulativeScalingIndex) {
     DEBUG_START_TIME();
 //    try {
         beagle::BeagleImpl* beagleInstance = beagle::getBeagleInstance(instance);
@@ -1313,7 +1455,7 @@ int beagleCopyScaleFactors(int instance,
     //    }
 }
 
-int beagleGetScaleFactors(int instance,                        
+int beagleGetScaleFactors(int instance,
                            int srcScalingIndex,
                            double* scaleFactors) {
     DEBUG_START_TIME();
@@ -1649,3 +1791,75 @@ int beagleGetSiteDerivatives(int instance,
     return returnValue;
 }
 
+int beagleCalculateEdgeDerivatives(int instance,
+                                   const int *postBufferIndices,
+                                   const int *preBufferIndices,
+                                   const int *derivativeMatrixIndices,
+                                   const int *categoryWeightsIndices,
+                                   int count,
+                                   double *outDerivatives,
+                                   double *outSumDerivatives,
+                                   double *outSumSquaredDerivatives) {
+    DEBUG_START_TIME();
+
+    beagle::BeagleImpl *beagleInstance = beagle::getBeagleInstance(instance);
+    if (beagleInstance == NULL) {
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    }
+
+    int returnValue = beagleInstance->calculateEdgeDerivatives(postBufferIndices,
+                                                               preBufferIndices,
+                                                               derivativeMatrixIndices,
+                                                               categoryWeightsIndices,
+                                                               NULL,
+                                                               NULL,
+                                                               count,
+                                                               outDerivatives,
+                                                               outSumDerivatives,
+                                                               outSumSquaredDerivatives);
+
+    DEBUG_END_TIME();
+
+    return returnValue;
+}
+
+int beagleCalculateCrossProductDerivative(int instance,
+                                   const int *postBufferIndices,
+                                   const int *preBufferIndices,
+                                   const int *categoryRatesIndices,
+                                   const int *categoryWeightsIndices,
+                                   const double *edgeLengths,
+                                   int count,
+                                   double *outSumDerivatives,
+                                   double *outSumSquaredDerivatives) {
+    DEBUG_START_TIME();
+
+    beagle::BeagleImpl *beagleInstance = beagle::getBeagleInstance(instance);
+    if (beagleInstance == NULL) {
+        return BEAGLE_ERROR_UNINITIALIZED_INSTANCE;
+    }
+
+    int returnValue = beagleInstance->calculateCrossProducts(postBufferIndices,
+                                                               preBufferIndices,
+                                                               categoryRatesIndices,
+                                                               categoryWeightsIndices,
+                                                               edgeLengths,
+                                                               count,
+                                                               outSumDerivatives,
+                                                               outSumSquaredDerivatives);
+
+    DEBUG_END_TIME();
+
+    return returnValue;
+}
+
+int beagleCalculateEdgeDerivative(int instance, const int *postBufferIndices, const int *preBufferIndices,
+                                  const int rootBufferIndex,
+                                  const int *firstDerivativeIndices, const int *secondDerivativeIndices,
+                                  const int categoryWeightsIndex, const int categoryRatesIndex,
+                                  const int stateFrequenciesIndex, const int *cumulativeScaleIndices, int count,
+                                  double *outFirstDerivative, double *outDiagonalSecondDerivative) {
+    fprintf(stderr, "Depricated");
+    return BEAGLE_ERROR_NO_IMPLEMENTATION;
+}
+
