@@ -5,25 +5,15 @@
  *
  * This file is part of BEAGLE.
  *
- * BEAGLE is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation, either version 3 of
- * the License, or (at your option) any later version.
- *
- * BEAGLE is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with BEAGLE.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * Use of this source code is governed by an MIT-style
+ * license that can be found in the LICENSE file or at
+ * https://opensource.org/licenses/MIT.
  *
  * @brief This file documents the API as well as header for the
  * Broad-platform Evolutionary Analysis General Likelihood Evaluator
  *
  * KEY CONCEPTS
- * 
+ *
  * The key to BEAGLE performance lies in delivering fine-scale
  * parallelization while minimizing data transfer and memory copy overhead.
  * To accomplish this, the library lacks the concept of data structure for
@@ -42,9 +32,9 @@
  * calls can be asynchronous, allowing the calling program to implement
  * other coarse-scale parallelization schemes such as evaluating
  * independent genes or running concurrent Markov chains.
- * 
+ *
  * USAGE
- * 
+ *
  * To use the library, a client program first creates an instance of BEAGLE
  * by calling beagleCreateInstance; multiple instances per client are
  * possible and encouraged. All additional functions are called with a
@@ -61,7 +51,7 @@
  * can then be specified, including the equilibrium state frequencies, the
  * rates for one or more substitution rate categories and their weights,
  * and finally, the eigen decomposition for the substitution process.
- * 
+ *
  * In order to calculate the likelihood of a particular tree, the client
  * program then specifies a series of integration operations that
  * correspond to steps in Felsenstein’s algorithm. Finite-time transition
@@ -77,7 +67,7 @@
  * software to keep track of these dependencies. The final step in
  * evaluating the phylogenetic model is done using an API call that yields
  * a single log likelihood for the model given the data.
- * 
+ *
  * Aspects of the BEAGLE API design support both maximum likelihood (ML)
  * and Bayesian phylogenetic tree inference. For ML inference, API calls
  * can calculate first and second derivatives of the likelihood with
@@ -102,7 +92,7 @@
  * @author Marc Suchard
  * @author David Swofford
  * @author Derrick Zwickl
- * 
+ *
  */
 
 #ifndef __beagle__
@@ -142,45 +132,48 @@ enum BeagleReturnCodes {
 enum BeagleFlags {
     BEAGLE_FLAG_PRECISION_SINGLE    = 1 << 0,    /**< Single precision computation */
     BEAGLE_FLAG_PRECISION_DOUBLE    = 1 << 1,    /**< Double precision computation */
-    
+
     BEAGLE_FLAG_COMPUTATION_SYNCH   = 1 << 2,    /**< Synchronous computation (blocking) */
     BEAGLE_FLAG_COMPUTATION_ASYNCH  = 1 << 3,    /**< Asynchronous computation (non-blocking) */
-    
+
     BEAGLE_FLAG_EIGEN_REAL          = 1 << 4,    /**< Real eigenvalue computation */
     BEAGLE_FLAG_EIGEN_COMPLEX       = 1 << 5,    /**< Complex eigenvalue computation */
-    
+
     BEAGLE_FLAG_SCALING_MANUAL      = 1 << 6,    /**< Manual scaling */
     BEAGLE_FLAG_SCALING_AUTO        = 1 << 7,    /**< Auto-scaling on (deprecated, may not work correctly) */
     BEAGLE_FLAG_SCALING_ALWAYS      = 1 << 8,    /**< Scale at every updatePartials (deprecated, may not work correctly) */
     BEAGLE_FLAG_SCALING_DYNAMIC     = 1 << 25,   /**< Manual scaling with dynamic checking (deprecated, may not work correctly) */
-    
+
     BEAGLE_FLAG_SCALERS_RAW         = 1 << 9,    /**< Save raw scalers */
     BEAGLE_FLAG_SCALERS_LOG         = 1 << 10,   /**< Save log scalers */
-    
+
     BEAGLE_FLAG_INVEVEC_STANDARD    = 1 << 20,   /**< Inverse eigen vectors passed to BEAGLE have not been transposed */
     BEAGLE_FLAG_INVEVEC_TRANSPOSED  = 1 << 21,   /**< Inverse eigen vectors passed to BEAGLE have been transposed */
-    
+
     BEAGLE_FLAG_VECTOR_SSE          = 1 << 11,   /**< SSE computation */
     BEAGLE_FLAG_VECTOR_AVX          = 1 << 24,   /**< AVX computation */
     BEAGLE_FLAG_VECTOR_NONE         = 1 << 12,   /**< No vector computation */
-    
+
     BEAGLE_FLAG_THREADING_CPP       = 1 << 30,   /**< C++11 threading */
     BEAGLE_FLAG_THREADING_OPENMP    = 1 << 13,   /**< OpenMP threading */
     BEAGLE_FLAG_THREADING_NONE      = 1 << 14,   /**< No threading (default) */
-    
+
     BEAGLE_FLAG_PROCESSOR_CPU       = 1 << 15,   /**< Use CPU as main processor */
     BEAGLE_FLAG_PROCESSOR_GPU       = 1 << 16,   /**< Use GPU as main processor */
     BEAGLE_FLAG_PROCESSOR_FPGA      = 1 << 17,   /**< Use FPGA as main processor */
     BEAGLE_FLAG_PROCESSOR_CELL      = 1 << 18,   /**< Use Cell as main processor */
     BEAGLE_FLAG_PROCESSOR_PHI       = 1 << 19,   /**< Use Intel Phi as main processor */
     BEAGLE_FLAG_PROCESSOR_OTHER     = 1 << 26,   /**< Use other type of processor */
-    
+
     BEAGLE_FLAG_FRAMEWORK_CUDA      = 1 << 22,   /**< Use CUDA implementation with GPU resources */
     BEAGLE_FLAG_FRAMEWORK_OPENCL    = 1 << 23,   /**< Use OpenCL implementation with GPU resources */
     BEAGLE_FLAG_FRAMEWORK_CPU       = 1 << 27,   /**< Use CPU implementation */
 
     BEAGLE_FLAG_PARALLELOPS_STREAMS = 1 << 28,   /**< Operations in updatePartials may be assigned to separate device streams */
-    BEAGLE_FLAG_PARALLELOPS_GRID    = 1 << 29    /**< Operations in updatePartials may be folded into single kernel launch (necessary for partitions; typically performs better for problems with fewer pattern sites) */
+    BEAGLE_FLAG_PARALLELOPS_GRID    = 1 << 29,   /**< Operations in updatePartials may be folded into single kernel launch (necessary for partitions; typically performs better for problems with fewer pattern sites) */
+
+    BEAGLE_FLAG_PREORDER_TRANSPOSE_MANUAL = 1 << 30, /**< Pre-order transition matrices passed to BEAGLE have been transposed */
+    BEAGLE_FLAG_PREORDER_TRANSPOSE_AUTO   = 1 << 31 /**< Automatically transpose pre-order transition matrices */
 };
 
 
@@ -315,7 +308,7 @@ BEAGLE_DLLEXPORT BeagleResourceList* beagleGetResourceList(void);
  * benchmark times and CPU performance ratios for each resource. Resources are benchmarked
  * with the given analysis parameters and the array is ordered from fastest to slowest.
  * If there is an error the function returns NULL.
- * 
+ *
  * @param tipCount              Number of tip data elements (input)
  * @param compactBufferCount    Number of compact state representation tips (input)
  * @param stateCount            Number of states in the continuous-time Markov chain (input)
@@ -328,14 +321,14 @@ BEAGLE_DLLEXPORT BeagleResourceList* beagleGetResourceList(void);
  *                               see BeagleFlags (input)
  * @param requirementFlags      Bit-flags indicating required implementation characteristics,
  *                               see BeagleFlags (input)
- * @param eigenModelCount       Number of full-alignment rate matrix eigen-decomposition models (input) 
+ * @param eigenModelCount       Number of full-alignment rate matrix eigen-decomposition models (input)
  * @param partitionCount        Number of partitions (input)
  * @param calculateDerivatives  Indicates if calculation of derivatives are required (input)
  * @param benchmarkFlags        Bit-flags indicating benchmarking preferences (input)
  *
- * @return An ordered (fastest to slowest) list of hardware resources available to the library as a 
+ * @return An ordered (fastest to slowest) list of hardware resources available to the library as a
  * BeagleBenchmarkedResourceList for the specified analysis parameters
- * 
+ *
  */
 BEAGLE_DLLEXPORT BeagleBenchmarkedResourceList* beagleGetBenchmarkedResourceList(int tipCount,
                                                     int compactBufferCount,
@@ -425,7 +418,7 @@ BEAGLE_DLLEXPORT int beagleFinalize(void);
  * implementation. It should only be called after beagleCreateInstance and requires the
  * BEAGLE_FLAG_THREADING_CPP flag to be set. It has no effect on GPU-based
  * implementations. It has no effect with the default BEAGLE_FLAG_THREADING_NONE setting.
- * If BEAGLE_FLAG_THREADING_CPP is set and this function is not called BEAGLE will use 
+ * If BEAGLE_FLAG_THREADING_CPP is set and this function is not called BEAGLE will use
  * a heuristic to set an appropriate number of threads.
  *
  * @param instance             Instance number (input)
@@ -471,12 +464,30 @@ BEAGLE_DLLEXPORT int beagleSetTipStates(int instance,
 BEAGLE_DLLEXPORT int beagleSetTipPartials(int instance,
                          int tipIndex,
                          const double* inPartials);
+/**
+ * @brief Set an instance pre-order partials buffer for root node
+ *
+ * This function copies an array of stateFrequencies into an instance buffer as the pre-order partials for the
+ * root node.
+ *
+ * @param instance                  Instance number (input)
+ * @param bufferIndices            List of partialsBuffer indices to set (input)
+ * @param stateFrequenciesIndices  List of state frequencies for each partialsBuffer (input). There
+ *                                  should be one set for each of parentBufferIndices
+ * @param count                    Number of partialsBuffer to integrate (input)
+ **
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleSetRootPrePartials(const int instance,
+                                              const int *bufferIndices,
+                                              const int *stateFrequenciesIndices,
+                                              int count);
 
 /**
  * @brief Set an instance partials buffer
  *
  * This function copies an array of partials into an instance buffer. The inPartials array should
- * be stateCount * patternCount * categoryCount in length. 
+ * be stateCount * patternCount * categoryCount in length.
  *
  * @param instance      Instance number in which to set a partialsBuffer (input)
  * @param bufferIndex   Index of destination partialsBuffer (input)
@@ -539,8 +550,8 @@ BEAGLE_DLLEXPORT int beagleSetEigenDecomposition(int instance,
  */
 BEAGLE_DLLEXPORT int beagleSetStateFrequencies(int instance,
                                          int stateFrequenciesIndex,
-                                         const double* inStateFrequencies);    
-    
+                                         const double* inStateFrequencies);
+
 /**
  * @brief Set a category weights buffer
  *
@@ -596,7 +607,7 @@ BEAGLE_DLLEXPORT int beagleSetCategoryRatesWithIndex(int instance,
  */
 BEAGLE_DLLEXPORT int beagleSetPatternWeights(int instance,
                                        const double* inPatternWeights);
-   
+
 /**
  * @brief Set pattern partition assignments
  *
@@ -639,11 +650,11 @@ BEAGLE_DLLEXPORT int beagleSetPatternPartitions(int instance,
  * This function convolves two lists of transition probability matrices.
  *
  * @param instance                  Instance number (input)
- * @param firstIndices              List of indices of the first transition probability matrices 
+ * @param firstIndices              List of indices of the first transition probability matrices
  *                                   to convolve (input)
  * @param secondIndices             List of indices of the second transition probability matrices
  *                                   to convolve (input)
- * @param resultIndices             List of indices of resulting transition probability matrices 
+ * @param resultIndices             List of indices of resulting transition probability matrices
  *                                   (input)
  * @param matrixCount               Length of lists
  */
@@ -653,6 +664,43 @@ BEAGLE_DLLEXPORT int beagleConvolveTransitionMatrices(int instance,
                                     const int* resultIndices,
                                     int matrixCount);
 
+/**
+ * @brief Adds lists of transition probability matrices
+ *
+ * This function adds two lists of transition probability matrices.
+ *
+ * @param instance                  Instance number (input)
+ * @param firstIndices              List of indices of the first transition probability matrices
+ *                                   to add (input)
+ * @param secondIndices             List of indices of the second transition probability matrices
+ *                                   to add (input)
+ * @param resultIndices             List of indices of resulting transition probability matrices
+ *                                   (input)
+ * @param matrixCount               Length of lists
+ */
+BEAGLE_DLLEXPORT int beagleAddTransitionMatrices(int instance,
+                                                 const int* firstIndices,
+                                                 const int* secondIndices,
+                                                 const int* resultIndices,
+                                                 int matrixCount);
+
+/**
+ * @brief Transpose lists of transition probability matrices
+ *
+ * This function transposes lists of transition probability matrices.
+ *
+ * @param instance                  Instance number (input)
+ * @param inputIndices             List of indices of transition probability matrices
+ *                                   to transpose (input)
+ * @param resultIndices             List of indices of resulting transition probability matrices
+ *                                   (input)
+ * @param matrixCount               Length of lists
+ */
+BEAGLE_DLLEXPORT int beagleTransposeTransitionMatrices(int instance,
+                                                       const int* inputIndices,
+                                                       const int* resultIndices,
+                                                       int matrixCount);
+
 /**
  * @brief Calculate a list of transition probability matrices
  *
@@ -680,6 +728,37 @@ BEAGLE_DLLEXPORT int beagleUpdateTransitionMatrices(int instance,
                                    const double* edgeLengths,
                                    int count);
 
+/**
+ * @brief Calculate a list of transition probability matrices with
+ *         each category using a different eigen decompsition
+ *
+ * This function calculates a list of transition probabilities matrices and their first and
+ * second derivatives (if requested). Each matrix is calculated for categoryCount categories,
+ * with each category using a different eigen decomposition based on eigenIndices.
+ *
+ * @param instance                  Instance number (input)
+ * @param eigenIndices              List of indices of eigen-decomposition buffers,
+ *                                   with categoryCount length (input)
+ * @param probabilityIndices        List of indices of transition probability matrices to update
+ *                                   (input)
+ * @param firstDerivativeIndices    List of indices of first derivative matrices to update
+ *                                   (input, NULL implies no calculation)
+ * @param secondDerivativeIndices    List of indices of second derivative matrices to update
+ *                                   (input, NULL implies no calculation)
+ * @param edgeLengths               List of edge lengths with which to perform calculations (input)
+ * @param count                     Length of lists (except for eigenIndices list)
+ *
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleUpdateTransitionMatricesWithModelCategories(int instance,
+                                   int* eigenIndices,
+                                   const int* probabilityIndices,
+                                   const int* firstDerivativeIndices,
+                                   const int* secondDerivativeIndices,
+                                   const double* edgeLengths,
+                                   int count);
+
+
 /**
  * @brief Calculate a list of transition probability matrices with multiple models
  *
@@ -729,6 +808,22 @@ BEAGLE_DLLEXPORT int beagleSetTransitionMatrix(int instance,
                               const double* inMatrix,
                               double paddedValue);
 
+/**
+ * @brief Set a differential transition probability matrix
+ *
+ * This function copies a differential transition probability matrix into a matrix buffer. The inMatrix array should be
+ * of size stateCount * stateCount * categoryCount and will contain one matrix for each rate category.
+ *
+ * @param instance      Instance number (input)
+ * @param matrixIndex   Index of matrix buffer (input)
+ * @param inMatrix      Pointer to source transition probability matrix (input)
+  *
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleSetDifferentialMatrix(int instance,
+                                                 int matrixIndex,
+                                                 const double* inMatrix);
+
 /**
  * @brief Get a finite-time transition probability matrix
  *
@@ -745,7 +840,7 @@ BEAGLE_DLLEXPORT int beagleSetTransitionMatrix(int instance,
 BEAGLE_DLLEXPORT int beagleGetTransitionMatrix(int instance,
                                 int matrixIndex,
                                 double* outMatrix);
-    
+
 /**
  * @brief Set multiple transition matrices
  *
@@ -768,7 +863,7 @@ BEAGLE_DLLEXPORT int beagleSetTransitionMatrices(int instance,
                                                  const double* paddedValues,
                                                  int count);
 
-    
+
 /**
  * @brief A list of integer indices which specify a partial likelihoods operation.
  */
@@ -797,9 +892,29 @@ typedef struct {
  * @return error code
  */
 BEAGLE_DLLEXPORT int beagleUpdatePartials(const int instance,
-                         const BeagleOperation* operations,
-                         int operationCount,
-                         int cumulativeScaleIndex);
+                                          const BeagleOperation* operations,
+                                          int operationCount,
+                                          int cumulativeScaleIndex);
+
+/**
+ * @brief Calculate or queue for calculation pre-order partials using a list of operations
+ *
+ * This function either calculates or queues for calculation a list pre-order partials. Implementations
+ * supporting ASYNCH may queue these calculations while other implementations perform these
+ * operations immediately and in order.
+ * ///TODO: consider > 3 degree nodes?
+ *
+ * @param instance                  Instance number (input)
+ * @param operations                BeagleOperation list specifying operations (input)
+ * @param operationCount            Number of operations (input)
+ * @param cumulativeScaleIndex      Index number of scaleBuffer to store accumulated factors (input)
+ *
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleUpdatePrePartials(const int instance,
+                                             const BeagleOperation* operations,
+                                             int operationCount,
+                                             int cumulativeScaleIndex);
 
 /**
  * @brief A list of integer indices which specify a partial likelihoods operation for a partitioned analysis.
@@ -833,6 +948,23 @@ BEAGLE_DLLEXPORT int beagleUpdatePartialsByPartition(const int instance,
                                                      const BeagleOperationByPartition* operations,
                                                      int operationCount);
 
+/**
+ * @brief Calculate or queue for calculation pre-order partials using a list of partition operations
+ *
+ * This function either calculates or queues for calculation a list partitioned pre-order partials. Implementations
+ * supporting ASYNCH may queue these calculations while other implementations perform these
+ * operations immediately and in order.
+ *
+ * @param instance                  Instance number (input)
+ * @param operations                BeagleOperation list specifying operations (input)
+ * @param operationCount            Number of operations (input)
+ *
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleUpdatePrePartialsByPartition(const int instance,
+                                                        const BeagleOperationByPartition* operations,
+                                                        int operationCount);
+
 /**
  * @brief Block until all calculations that write to the specified partials have completed.
  *
@@ -962,7 +1094,7 @@ BEAGLE_DLLEXPORT int beagleResetScaleFactorsByPartition(int instance,
 BEAGLE_DLLEXPORT int beagleCopyScaleFactors(int instance,
                                             int destScalingIndex,
                                             int srcScalingIndex);
-                                            
+
 /**
  * @brief Get scale factors
  *
@@ -972,10 +1104,10 @@ BEAGLE_DLLEXPORT int beagleCopyScaleFactors(int instance,
  * @param srcScalingIndex           Source scaleBuffer (input)
  * @param outScaleFactors           Pointer to which to receive scaleFactors (output)
  */
-BEAGLE_DLLEXPORT int beagleGetScaleFactors(int instance,                                            
+BEAGLE_DLLEXPORT int beagleGetScaleFactors(int instance,
                                            int srcScalingIndex,
-                                           double* outScaleFactors);     
-                                                                                                                                                                                                   
+                                           double* outScaleFactors);
+
 /**
  * @brief Calculate site log likelihoods at a root node
  *
@@ -1022,8 +1154,8 @@ BEAGLE_DLLEXPORT int beagleCalculateRootLogLikelihoods(int instance,
  * @param cumulativeScaleIndices   List of scaleBuffers containing accumulated factors to apply to
  *                                  each partialsBuffer (input). There should be one index for each
  *                                  of bufferIndices
- * @param partitionIndices         List of partition indices indicating which sites in each 
- *                                  partialsBuffer should be used (input). There should be one 
+ * @param partitionIndices         List of partition indices indicating which sites in each
+ *                                  partialsBuffer should be used (input). There should be one
  *                                  index for each of bufferIndices
  * @param partitionCount           Number of distinct partitionIndices (input)
  * @param count                    Number of sets of partitions to integrate across (input)
@@ -1043,6 +1175,66 @@ BEAGLE_DLLEXPORT int beagleCalculateRootLogLikelihoodsByPartition(int instance,
                                                                   double* outSumLogLikelihoodByPartition,
                                                                   double* outSumLogLikelihood);
 
+/**
+ * @brief Calculate site log likelihoods and derivatives along all edges
+ *
+ * This function integrates a list of partials at a parent and child node with respect
+ * to a set of partials-weights and state frequencies to return the log likelihood
+ * and first and second derivative sums
+ *
+ * @param instance                  Instance number (input)
+ * @param postBufferIndices               List of indices of post-order partialsBuffers (input)
+ * @param preBufferIndices                List of indices of pre-order partialsBuffers (input)
+ * @param rootBufferIndices               List of indices of root post-order partialsBuffers (input)
+ * @param insMatrixIndices                List indices of instantaneous transition matrices (input)
+ * @param firstDerivativeIndices          List indices of first derivative matrices (input)
+ * @param secondDerivativeIndices         List indices of second derivative matrices (input, turned off by NULL input)
+ * @param categoryWeightsIndices          List of weights to apply to each partialsBuffer (input)
+ * @param stateFrequenciesIndices         List of state frequencies for each partialsBuffer (input). There
+ *                                         should be one set for each of parentBufferIndices
+ * @param cumulativeScaleIndices          List of scaleBuffers containing accumulated factors to apply to
+ *                                         each partialsBuffer (input). There should be one index for each
+ *                                         of parentBufferIndices
+ * @param count                           Number of partialsBuffers (input)
+ * @param outFirstDerivative              Pointer to destination for resulting first derivative (output)
+ * @param outDiagonalSecondDerivative     Pointer to destination for resulting first derivative (output)
+ *
+ * @return error code
+ */
+BEAGLE_DLLEXPORT int beagleCalculateEdgeDerivative(int instance,
+                                                   const int *postBufferIndices,
+                                                   const int *preBufferIndices,
+                                                   const int rootBufferIndex,
+                                                   const int *firstDerivativeIndices,
+                                                   const int *secondDerivativeIndices,
+                                                   const int categoryWeightsIndex,
+                                                   const int categoryRatesIndex,
+                                                   const int stateFrequenciesIndex,
+                                                   const int *cumulativeScaleIndices,
+                                                   int count,
+                                                   double *outFirstDerivative,
+                                                   double *outDiagonalSecondDerivative);
+
+BEAGLE_DLLEXPORT int beagleCalculateEdgeDerivatives(int instance,
+                                                    const int *postBufferIndices,
+                                                    const int *preBufferIndices,
+                                                    const int *derivativeMatrixIndices,
+                                                    const int *categoryWeightsIndices,
+                                                    int count,
+                                                    double *outDerivatives,
+                                                    double *outSumDerivatives,
+                                                    double *outSumSquaredDerivatives);
+
+BEAGLE_DLLEXPORT int beagleCalculateCrossProductDerivative(int instance,
+                                                  const int *postBufferIndices,
+                                                  const int *preBufferIndices,
+                                                  const int *categoryRateIndices,
+                                                  const int *categoryWeightsIndices,
+                                                  const double *edgeLengths,
+                                                  int count,
+                                                  double *outSumDerivatives,
+                                                  double *outSumSquaredDerivatives);
+
 /**
  * @brief Calculate site log likelihoods and derivatives along an edge
  *
@@ -1085,7 +1277,7 @@ BEAGLE_DLLEXPORT int beagleCalculateEdgeLogLikelihoods(int instance,
                                       double* outSumSecondDerivative);
 
 /**
- * @brief Calculate multiple site log likelihoods and derivatives along an edge with 
+ * @brief Calculate multiple site log likelihoods and derivatives along an edge with
  *         per partition buffers
  *
  * This function integrates lists of partials at a parent and child node with respect
@@ -1105,8 +1297,8 @@ BEAGLE_DLLEXPORT int beagleCalculateEdgeLogLikelihoods(int instance,
  * @param cumulativeScaleIndices    List of scaleBuffers containing accumulated factors to apply to
  *                                   each partialsBuffer (input). There should be one index for each
  *                                   of parentBufferIndices
- * @param partitionIndices          List of partition indices indicating which sites in each 
- *                                   partialsBuffer should be used (input). There should be one 
+ * @param partitionIndices          List of partition indices indicating which sites in each
+ *                                   partialsBuffer should be used (input). There should be one
  *                                   index for each of parentBufferIndices
  * @param partitionCount            Number of distinct partitionIndices (input)
  * @param count                     Number of sets of partitions to integrate across (input)
@@ -1176,12 +1368,12 @@ BEAGLE_DLLEXPORT int beagleGetLogLikelihood(int instance,
 BEAGLE_DLLEXPORT int beagleGetDerivatives(int instance,
                                           double* outSumFirstDerivative,
                                           double* outSumSecondDerivative);
-                                          
+
 /**
  * @brief Get site log likelihoods for last beagleCalculateRootLogLikelihoods or
  *         beagleCalculateEdgeLogLikelihoods call
  *
- * This function returns the log likelihoods for each site 
+ * This function returns the log likelihoods for each site
  *
  * @param instance               Instance number (input)
  * @param outLogLikelihoods      Pointer to destination for resulting log likelihoods (output)
@@ -1194,7 +1386,7 @@ BEAGLE_DLLEXPORT int beagleGetSiteLogLikelihoods(int instance,
 /**
  * @brief Get site derivatives for last beagleCalculateEdgeLogLikelihoods call
  *
- * This function returns the derivatives for each site 
+ * This function returns the derivatives for each site
  *
  * @param instance               Instance number (input)
  * @param outFirstDerivatives    Pointer to destination for resulting first derivatives (output)
@@ -1204,8 +1396,8 @@ BEAGLE_DLLEXPORT int beagleGetSiteLogLikelihoods(int instance,
  */
 BEAGLE_DLLEXPORT int beagleGetSiteDerivatives(int instance,
                                     double* outFirstDerivatives,
-                                    double* outSecondDerivatives);    
-    
+                                    double* outSecondDerivatives);
+
 /* using C calling conventions so that C programs can successfully link the beagle library
  * (closing brace)
  */
